/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.binding.http.rt;

import javax.xml.namespace.QName;

/**
 * A 'catch-all' HTTPFault indicating fault is due to 'reasons attributable to the processing
 * of the message rather than to the contents of the message itself.'
 */

public class HTTPReceiverFault extends HTTPFault {

    private static final long serialVersionUID = 1L;

    public static final String DEFAULT_REASON = "An error occurred while processing message";

    /**
     * Constructs a <code>HTTPReceiverFault</code>.
     * 
     * @param faultReason
     */
    public HTTPReceiverFault(String faultReason) {
        super(RECEIVER_ERROR, faultReason);
    }

    /**
     * Constructs a <code>HTTPReceiverFault</code> .
     * 
     * @param faultReason
     * @param th
     *        Cause of fault. Note that stack trace will be embedded within fault detail. Use
     *        alternative constructor {@link #HTTPReceiverFault(HTTPVersion, String)} if this
     *        behavior is not desired.
     */
    public HTTPReceiverFault(String faultReason, Throwable th) {
        super(RECEIVER_ERROR, faultReason, th);
    }

    /**
     * Constructor used by subclasses which provide a more specific indication of fault.
     * 
     * @param soapVersion
     * @param faultName
     * @param faultReason
     */
    protected HTTPReceiverFault(QName faultName, String faultReason) {
        super(faultName, faultReason);
    }

    /**
     * Constructs a <code>HTTPReceiverFault</code> with implied reason
     * {@link #DEFAULT_REASON} and a provided cause.
     * 
     * @param th
     *        Cause of fault. Note that stack trace will be embedded within fault detail. Use
     *        alternative constructor {@link #HTTPReceiverFault(HTTPVersion, String)} if this
     *        behavior is not wanted.
     */
    public HTTPReceiverFault(Throwable th) {
        super(RECEIVER_ERROR, DEFAULT_REASON, th);
    }
    
}
