/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.binding.soap;

import java.util.Collection;
import java.util.Collections;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction.Direction;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HTTPHeader;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HTTPLocationInfo;

/**
 * Models an <code>UnderlyingProtocol</code> component which statically binds SOAP 1.1 to the HTTP
 * transmission protocol.
 * 
 * @see http://www.w3.org/TR/wsdl20-soap11-binding/
 */

public final class SOAP11HTTPProtocol extends SOAPHTTPProtocol {

    public static final String TYPE = "http://www.w3.org/2006/01/soap11/bindings/HTTP";

    public SOAP11HTTPProtocol() {
        super(TYPE);
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HyperTextTransferProtocol#getFaultHTTPContentEncoding(javax.xml.namespace.QName)
     */
    public String getFaultHTTPContentEncoding(QName faultName) {
        return null;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HyperTextTransferProtocol#getFaultHTTPHeaders(javax.xml.namespace.QName)
     */
    @SuppressWarnings("unchecked")
    public Collection<HTTPHeader> getFaultHTTPHeaders(QName faultName) {
        return Collections.EMPTY_SET;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HyperTextTransferProtocol#getHTTPContentEncodingDefault()
     */
    public String getHTTPContentEncodingDefault() {
        return null;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HyperTextTransferProtocol#getHTTPLocationInfo(javax.xml.namespace.QName)
     */
    public HTTPLocationInfo getHTTPLocationInfo(QName operationName) {
        return null;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HyperTextTransferProtocol#getHTTPQueryParamaterSeparator(javax.xml.namespace.QName)
     */
    public String getHTTPQueryParameterSeparator(QName operationName) {
        return null;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HyperTextTransferProtocol#getHTTPQueryParameterSeparatorDefault()
     */
    public String getHTTPQueryParameterSeparatorDefault() {
        return null;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HyperTextTransferProtocol#getMessageReferenceHTTPContentEncoding(javax.xml.namespace.QName,
     *      org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction.Direction)
     */
    public String getMessageReferenceHTTPContentEncoding(QName operationName, Direction direction) {
        return null;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HyperTextTransferProtocol#getMessageReferenceHTTPHeaders(javax.xml.namespace.QName,
     *      org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction.Direction)
     */
    @SuppressWarnings("unchecked")
    public Collection<HTTPHeader> getMessageReferenceHTTPHeaders(QName operationName, Direction direction) {
        return Collections.EMPTY_SET;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HyperTextTransferProtocol#getOperationHTTPContentEncoding(javax.xml.namespace.QName)
     */
    public String getOperationHTTPContentEncoding(QName operationName) {
        return null;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HyperTextTransferProtocol#useHTTPCookies()
     */
    public Boolean useHTTPCookies() {
        return null;
    }

    protected void validate() throws DeploymentException {
        if (!(parent instanceof SOAP11Binding)) {
            throw new DeploymentException("Parent not instanceof " + SOAP11Binding.class);
        }
    }

    // *********************** Identity Methods ************************************

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("type: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(TYPE);
        return sb.toString();
    }

    @Override
    public int hashCode() {
        return hashCode(TYPE);
    }

    @Override
    public boolean equals(Object obj) {
        if (obj instanceof SOAP11HTTPProtocol) {
            return true;
        } else {
            return false;
        }
    }

    private int hashCode(Object object) {
        return (object == null ? 0 : object.hashCode());
    }

}
