/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.binding.soap;

import java.net.InetAddress;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.UnknownHostException;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.ext.feature.http.server.HTTPServerFeature;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HTTPLocationInfo;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HTTPMethod;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HyperTextTransferProtocol;
import org.bluestemsoftware.specification.eoa.system.SystemContext;
import org.bluestemsoftware.specification.eoa.system.server.Server;

public abstract class SOAPHTTPProtocol extends SOAPUnderlyingProtocol implements HyperTextTransferProtocol {

    public SOAPHTTPProtocol(String protocolType) {
        super(protocolType);
    }

    protected abstract void validate() throws DeploymentException;

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HyperTextTransferProtocol#createEndpointAddress(java.lang.String, java.lang.String, java.lang.String)
     */
    public String createEndpointAddress(String path, String query, String fragment) throws DeploymentException {
        String hostName = null;
        try {
            hostName = InetAddress.getLocalHost().getHostName();
        } catch (UnknownHostException ue) {
            throw new DeploymentException("Error creating endpoint address." + ue);
        }
        Server eoaServer = SystemContext.getContext().getSystem().getServer();
        HTTPServerFeature httpServer = eoaServer.getFeature(HTTPServerFeature.class);
        if (httpServer == null) {
            throw new DeploymentException("Error creating HTTP endpoint address. Feature "
                    + HTTPServerFeature.class
                    + " is undefined.");
        }
        int port = 0;
        int[] ports = httpServer.getPorts(false);
        if (ports.length == 0) {
            ports = httpServer.getPorts(true);
            if (ports.length == 0) {
                throw new DeploymentException("Error creating HTTP endpoint address. Feature "
                        + HTTPServerFeature.class
                        + " defines no ports.");
            }
        } else {
            port = ports[0];
        }
        // if server is listening on more than one port, the choice
        // here is arbitrary, i.e. if a specific port is required
        // then user must define address, e.g. override callback
        // address generated by ws-addressing module, etc ... w/
        // address that uses specific port
        try {
            return new URI("http", null, hostName, port, path, query, fragment).toString();
        } catch (URISyntaxException ue) {
            throw new DeploymentException("Error creating HTTP endpoint address. " + ue);
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HyperTextTransferProtocol#getFaultSerialization(javax.xml.namespace.QName)
     */
    public String getFaultSerialization(QName operationName) {
        return null; // not used with SOAP binding
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HyperTextTransferProtocol#getInputSerialization(javax.xml.namespace.QName)
     */
    public String getInputSerialization(QName operationName) {
        return null; // not used with SOAP binding
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HyperTextTransferProtocol#getMethod(javax.xml.namespace.QName)
     */
    public HTTPMethod getMethod(QName operationName) {
        return null; // not used with SOAP binding
    }
    
    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HyperTextTransferProtocol#getErrorStatusCode(javax.xml.namespace.QName)
     */
    public Integer getErrorStatusCode(QName faultName) {
        return null; // not used with SOAP binding
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HyperTextTransferProtocol#getMethodDefault()
     */
    public HTTPMethod getMethodDefault() {
        return null; // not used with SOAP binding
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HyperTextTransferProtocol#getOutputSerialization(javax.xml.namespace.QName)
     */
    public String getOutputSerialization(QName operationName) {
        return null; // not used with SOAP binding
    }
    
    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPHTTPProtocol#getHTTPLocationInfo(java.lang.String)
     */
    public HTTPLocationInfo getHTTPLocationInfo(String location) {
        return null;  // not used with SOAP binding
    }

}
