/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.binding.soap;

import java.util.Collection;
import java.util.Set;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.component.Component;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.appdata.soap.ApplicationDataFeature;
import org.bluestemsoftware.specification.eoa.system.server.ModulableFeature;

/**
 * Abstractly describes a concrete message, i.e. an <code>InterfaceMessage</code> bound to
 * SOAP messaging protocol.
 */
public interface SOAPMessage extends Component {

    /**
     * Indicates message style, i.e. 'document style' or 'rpc style.'
     * 
     * @return SOAPMessageStyle
     */
    public SOAPMessageStyle getMessageStyle();

    /**
     * Returns collection of metadata which abstractly describes expected soap:header content.
     * <p>
     * Note that <code>SOAPHeaderBlock</code> definitions SHOULD NOT be used to constrain
     * instances of {@link ModulableFeature} other than {@link ApplicationDataFeature}. All
     * other features SHOULD be constrained via policy.
     * 
     * @return Set of declared header blocks, or an empty Set if none declared.
     */
    public Collection<SOAPHeaderBlock> getHeaders();

    /**
     * Convenience method which returns names of all headers defined as required by this
     * concrete message definition.
     * 
     * @return a <code>Set</code> of qualified names, each of which identifies a required
     *         header, or <code>null</code> if no headers required.
     */
    public Set<QName> getRequiredHeaders();

    /**
     * Gets a <code>SOAPHeaderBlock</code> by qualified name.
     * 
     * @param name
     * @return SOAPHeaderBlock or <code>null</code> if undefined.
     */
    public SOAPHeaderBlock getHeader(QName name);

    /*
     * (non-Javadoc)
     */
    void setParent(Component parent);

}
