/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.binding.soap;

/**
 * An eumeration of available SOAP version specifications.
 */
public enum SOAPVersion {
    
    
    SOAP11, SOAP12;

    public static final String SOAP_11_ENVELOPE_NS = "http://schemas.xmlsoap.org/soap/envelope/";
    public static final String SOAP_12_ENVELOPE_NS = "http://www.w3.org/2003/05/soap-envelope";

    /**
     * Translates soap:Envelope namespace into an enumerated value.
     * 
     * @param envelopeNamespace
     * @return
     */
    public static SOAPVersion valueOfEnvelopeNamespace(String envelopeNamespace) {
        if (envelopeNamespace == null) {
            throw new IllegalArgumentException("envelopeNamespace null");
        }
        if (envelopeNamespace.equals(SOAP_11_ENVELOPE_NS)) {
            return SOAP11;
        }
        if (envelopeNamespace.equals(SOAP_12_ENVELOPE_NS)) {
            return SOAP12;
        }
        throw new IllegalArgumentException("namespace " + envelopeNamespace + "is undefined.");
    }
    
    /**
     * Translates an instance of <code>SOAPVersion</code> into an soap:Envelope namespace.
     * 
     * @param soapVersion
     * @return
     */
    public static String toEnvelopeNamespace(SOAPVersion soapVersion) {
        if (soapVersion == null) {
            throw new IllegalArgumentException("soapVersion null");
        }
        if (soapVersion == SOAPVersion.SOAP11) {
            return SOAP_11_ENVELOPE_NS;
        } else {
            return SOAP_12_ENVELOPE_NS;
        }
    }

}
