/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.binding.soap.rt;

import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentName;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointReference;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction.Direction;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.ActionContext;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.SystemFault;
import org.bluestemsoftware.specification.eoa.ext.binding.BindingFactory;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.vm.rt.VMTransportProtocolRT;
import org.w3c.dom.Element;

public abstract class SOAPVMBindingRT extends SOAPBindingRT implements VMTransportProtocolRT {

    /*
     * (non-Javadoc)
     * 
     * SPI
     */
    public interface Provider extends SOAPBindingRT.Provider {
        
        public void spi_sendAction(EndpointReference endpointReference, ActionContext actionContext) throws SystemFault;

        public void spi_receiveAction(EndpointReference endpointReference, Direction direction, Element envelope);
        
    }

    public SOAPVMBindingRT(BindingFactory factory, Provider provider, ComponentName componentName,
            String protocolType) {
        super(factory, provider, componentName, protocolType);
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.binding.rt.BindingRT#sendAction(org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointReference,
     *      org.bluestemsoftware.specification.eoa.component.intrface.rt.ActionContext)
     */
    @Override
    public void sendAction(EndpointReference endpointReference, ActionContext actionContext) throws SystemFault {
        if (provider == null) {
            return;
        }
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            ((Provider)provider).spi_sendAction(endpointReference, actionContext);
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.vm.rt.VMTransportProtocolRT#receiveAction(org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointReference, org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction.Direction, org.w3c.dom.Element)
     */
    public void receiveAction(EndpointReference endpointReference, Direction direction, Element envelope) {
        if (provider == null) {
            return;
        }
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            ((Provider)provider).spi_receiveAction(endpointReference, direction, envelope);
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

}
