/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.binding.soap.rt;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPVersion;

public class SOAPVersionMismatchFault extends SOAPSenderFault {

    private static final long serialVersionUID = 1L;

    public static final QName SOAP_11_FAULT_CODE = new QName("http://schemas.xmlsoap.org/soap/envelope/",
            "VersionMismatch", SOAP_NS_PREFIX);
    public static final QName SOAP_12_FAULT_CODE = new QName("http://www.w3.org/2003/05/soap-envelope",
            "VersionMismatch", SOAP_NS_PREFIX);

    public SOAPVersionMismatchFault(SOAPVersion soapVersion) {
        super(soapVersion, "Version Mismatch");
    }

    /**
     * In this case soapVersion represents the received version, i.e. which is version of fault
     * envelope which should be returned to sender. Note that receiver SHOULD return an upgrade
     * soap header block where appropriate.
     */
    @Override
    public QName getFaultCode() {
        if (soapVersion == SOAPVersion.SOAP11) {
            return SOAP_11_FAULT_CODE;
        } else {
            return SOAP_12_FAULT_CODE;
        }
    }

}
