/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.container.standalone10;

import java.util.Properties;
import java.util.Set;

public class StandaloneContainerConfiguration {
    
    /*
     * default values. do not change these, without updating default
     * values defined within configuration schema.
     */
    
    public static final int DEFAULT_TRANSACTION_TIMEOUT = 60 * 5; // 5 minutes
    public static final int DEFAULT_MIN_POOL_SIZE = 1;
    public static final int DEFAULT_MAX_POOL_SIZE = 10;
    public static final int DEFAULT_BLOCKING_TIMEOUT_MILLISECONDS = 60000; // 1 minute
    public static final int DEFAULT_IDLE_TIMEOUT_MILLISECONDS = 15 * 60 * 1000; // 15 minutes
    public static final boolean DEFAULT_MATCH_ONE = false;
    public static final boolean DEFAULT_MATCH_ALL = false;
    public static final boolean DEFAULT_SELECT_ONE_ASSUME_MATCH = true;

    private TransactionSupportInfo transactionSupportInfo;
    private Set<ConnectorInfo> connectorInfos;
    
    public static enum PoolingSupportType {
        NONE, SINGLE, PARTITIONED
    }

    public static enum TransactionSupportType {
        NONE, LOCAL, XA
    }

    public static class TransactionSupportInfo {

        private int defaultTransactionTimeoutSeconds = DEFAULT_TRANSACTION_TIMEOUT;
        
        public int getDefaultTransactionTimeoutSeconds() {
            return defaultTransactionTimeoutSeconds;
        }

        public void setDefaultTransactionTimeoutSeconds(int defaultTransactionTimeoutSeconds) {
            this.defaultTransactionTimeoutSeconds = defaultTransactionTimeoutSeconds;
        }
        
    }

    public static class ConnectorInfo {

        private String type;
        private String impl;
        private String connectorName;
        private Properties properties;
        private ConnectorTransactionSupportInfo transactionSupportInfo;
        private PoolingSupportInfo poolingSupportInfo;

        public static class ConnectorTransactionSupportInfo {

            private boolean useTransactionCaching = true; //default
            private boolean useThreadCaching = true; // default
            private TransactionSupportType transactionSupportType;

            public TransactionSupportType getTransactionSupportType() {
                return transactionSupportType;
            }

            public void setTransactionSupportType(TransactionSupportType transactionSupportType) {
                this.transactionSupportType = transactionSupportType;
            }

            public boolean isUseThreadCaching() {
                return useThreadCaching;
            }

            public void setUseThreadCaching(boolean useThreadCaching) {
                this.useThreadCaching = useThreadCaching;
            }

            public boolean isUseTransactionCaching() {
                return useTransactionCaching;
            }

            public void setUseTransactionCaching(boolean useTransactionCaching) {
                this.useTransactionCaching = useTransactionCaching;
            }

        }

        public static class PoolingSupportInfo {

            private int minPoolSize = DEFAULT_MIN_POOL_SIZE;
            private int maxPoolSize = DEFAULT_MAX_POOL_SIZE;
            private int blockingTimeoutMilliseconds = DEFAULT_BLOCKING_TIMEOUT_MILLISECONDS;
            private int idleTimeoutMilliseconds = DEFAULT_IDLE_TIMEOUT_MILLISECONDS;
            private boolean matchOne = DEFAULT_MATCH_ONE;
            private boolean matchAll = DEFAULT_MATCH_ALL;
            private boolean selectOneAssumeMatch = DEFAULT_SELECT_ONE_ASSUME_MATCH;
            private boolean partitionByConnectionRequest;
            private boolean partitionBySubject;
            private PoolingSupportType poolingSupportType;

            public int getBlockingTimeoutMilliseconds() {
                return blockingTimeoutMilliseconds;
            }

            public void setBlockingTimeoutMilliseconds(int blockingTimeoutMilliseconds) {
                this.blockingTimeoutMilliseconds = blockingTimeoutMilliseconds;
            }

            public int getIdleTimeoutMilliseconds() {
                return idleTimeoutMilliseconds;
            }

            public void setIdleTimeoutMilliseconds(int idleTimeoutMilliseconds) {
                this.idleTimeoutMilliseconds = idleTimeoutMilliseconds;
            }

            public boolean isMatchAll() {
                return matchAll;
            }

            public void setMatchAll(boolean matchAll) {
                this.matchAll = matchAll;
            }

            public boolean isMatchOne() {
                return matchOne;
            }

            public void setMatchOne(boolean matchOne) {
                this.matchOne = matchOne;
            }

            public int getMaxPoolSize() {
                return maxPoolSize;
            }

            public void setMaxPoolSize(int maxPoolSize) {
                this.maxPoolSize = maxPoolSize;
            }

            public int getMinPoolSize() {
                return minPoolSize;
            }

            public void setMinPoolSize(int minPoolSize) {
                this.minPoolSize = minPoolSize;
            }

            public boolean isPartitionByConnectionRequest() {
                return partitionByConnectionRequest;
            }

            public void setPartitionByConnectionRequest(boolean partitionByConnectionRequest) {
                this.partitionByConnectionRequest = partitionByConnectionRequest;
            }

            public boolean isPartitionBySubject() {
                return partitionBySubject;
            }

            public void setPartitionBySubject(boolean partitionBySubject) {
                this.partitionBySubject = partitionBySubject;
            }

            public PoolingSupportType getPoolingSupportType() {
                return poolingSupportType;
            }

            public void setPoolingSupportType(PoolingSupportType poolingSupportType) {
                this.poolingSupportType = poolingSupportType;
            }

            public boolean isSelectOneAssumeMatch() {
                return selectOneAssumeMatch;
            }

            public void setSelectOneAssumeMatch(boolean selectOneAssumeMatch) {
                this.selectOneAssumeMatch = selectOneAssumeMatch;
            }

        }

        public String getConnectorName() {
            return connectorName;
        }

        public void setConnectorName(String connectorName) {
            this.connectorName = connectorName;
        }

        public String getImpl() {
            return impl;
        }

        public void setImpl(String impl) {
            this.impl = impl;
        }

        public PoolingSupportInfo getPoolingSupportInfo() {
            return poolingSupportInfo;
        }

        public void setPoolingSupportInfo(PoolingSupportInfo poolingSupportInfo) {
            this.poolingSupportInfo = poolingSupportInfo;
        }

        public Properties getProperties() {
            return properties;
        }

        public void setProperties(Properties properties) {
            this.properties = properties;
        }

        public ConnectorTransactionSupportInfo getTransactionSupportInfo() {
            return transactionSupportInfo;
        }

        public void setTransactionSupportInfo(ConnectorTransactionSupportInfo transactionSupportInfo) {
            this.transactionSupportInfo = transactionSupportInfo;
        }

        public String getType() {
            return type;
        }

        public void setType(String type) {
            this.type = type;
        }

    }

    public Set<ConnectorInfo> getConnectorInfos() {
        return connectorInfos;
    }

    public void setConnectorInfos(Set<ConnectorInfo> connectorInfos) {
        this.connectorInfos = connectorInfos;
    }

    public TransactionSupportInfo getTransactionSupportInfo() {
        return transactionSupportInfo;
    }

    public void setTransactionSupportInfo(TransactionSupportInfo transactionSupportInfo) {
        this.transactionSupportInfo = transactionSupportInfo;
    }

}
