/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.feature.db.server;

import javax.sql.DataSource;

import org.bluestemsoftware.specification.eoa.system.container.Container;
import org.bluestemsoftware.specification.eoa.system.server.Feature;
import org.bluestemsoftware.specification.eoa.system.server.Server;

/**
 * An 'embedded' Database Server, i.e. a Database Server which runs within same jvm as
 * {@link Server}.
 */
public class DBServerFeature extends Feature {

    public static final String TYPE = DBServerFeature.class.getName();

    /*
     * (non-Javadoc)
     * 
     * SPI
     */
    public interface Provider extends Feature.Provider {
        public DataSource spi_getEmbeddedDataSource(boolean isXA);
    }
    
    public DBServerFeature(DBServerFeatureFactory factory, Feature.Provider provider) {
        super(factory, provider);
    }
    
    /**
     * Returns a <code>DataSource</code> that references an in-memory representation of
     * database server, i.e. a class scoped to this deployment. Used by
     * <code>ManagedConnectionFactory</code> to create 'managed' connections to embedded
     * server.
     * <p>
     * Note, this method SHOULD NOT be used to create user connections to embedded server.
     * Server SHOULD be accessed via 'managed' connections configured by container and
     * retreived via {@link Container#getConnector(String)}.
     * 
     * @param isXA
     *        indicates whether or not returned <code>DataSource</code> creates connections
     *        which are capable of participating in a distributed transaction. If false,
     *        connections supports local transactions only.
     * 
     * @return DataSource
     */
    public DataSource getEmbeddedDataSource(boolean isXA) {
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            return ((Provider)provider).spi_getEmbeddedDataSource(isXA);
        } finally {
            thread.setContextClassLoader(cl);
        }
    }
}
