/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.feature.http.client;

import java.io.IOException;

/**
 * An <code>HTTPClient</code> object used to connect to host(s) and send requests using HTTP
 * transmission protocol. Client MUST be HTTP version agnostic. Implementations MUST be
 * threadsafe and SHOULD be re-used.
 * <p>
 * Note that it is responsibility of implementer to manage setting of thread context
 * classloader as appropriate.
 */
public interface HTTPClient {

    /**
     * Sends an <code>HTTPClientRequest</code> and returns an <code>HTTPClientResponse</code>.
     * 
     * @param request
     *        http request.
     * @return Response http response.
     * @throws IOException
     *         if an error occurs sending request.
     */
    public HTTPClientResponse doSend(HTTPClientRequest request) throws IOException;

}
