/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.feature.http.server;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.security.Principal;
import java.util.Enumeration;
import java.util.Locale;
import java.util.Map;

/**
 * Models a request handled by the 'receiving' node.
 */
public interface HTTPServerRequest {

    // design note: this interface was created to break the dependency
    // our system-test artifact had on the javax servlet spec artifact
    // which effectively required every project to define the servlet
    // spec as a dependency
    
    public Object getAttribute(String name);

    public Enumeration<String> getAttributeNames();

    public String getAuthType();

    public String getCharacterEncoding();

    public int getContentLength();

    public String getContentType();

    public String getContextPath();

    public long getDateHeader(String name);

    public String getHeader(String name);

    public Enumeration<String> getHeaderNames();

    public Enumeration<String> getHeaders(String name);

    public InputStream getInputStream() throws IOException;

    public int getIntHeader(String name);

    public String getLocalAddr();

    public Locale getLocale();

    public Enumeration<Locale> getLocales();

    public String getLocalName();

    public int getLocalPort();

    public String getMethod();

    public String getParameter(String name);

    public Map<String, String[]> getParameterMap();

    public Enumeration<String> getParameterNames();

    public String[] getParameterValues(String name);

    public String getPathInfo();

    public String getPathTranslated();

    public String getProtocol();

    public String getQueryString();

    public BufferedReader getReader() throws IOException;

    public String getRemoteAddr();

    public String getRemoteHost();

    public int getRemotePort();

    public String getRemoteUser();

    public String getRequestedSessionId();

    public String getRequestURI();

    public StringBuffer getRequestURL();

    public String getScheme();

    public String getServerName();

    public int getServerPort();

    public String getServletPath();

    public Principal getUserPrincipal();

    public boolean isSecure();

    public boolean isUserInRole(String role);

    public void removeAttribute(String name);

    public void setAttribute(String name, Object o);

    public void setCharacterEncoding(String env) throws UnsupportedEncodingException;

}
