/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.feature.ws.mex;

import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.WSDLFactory;
import org.bluestemsoftware.specification.eoa.system.server.Feature;
import org.xml.sax.InputSource;

/**
 * An EOA <code>Feature</code> component which implements WS-MetadataExchange behavior in
 * conjunction with a ws-mex <code>Engine</code> component, i.e. for sending/receiving SOAP
 * messages. Feature SHOULD be used by {@link WSDLFactory} providers, i.e. when reading a
 * definition that references remotely defined metadata documents, e.g. wsdl, policy and
 * schema.
 * <p>
 * Note that feature SHOULD provide an optional configuration which provides ability to
 * exchange metadata as plain xml rather than as a SOAP message, via an extensible set of
 * transports, e.g. http.
 */
public class WSMexFeature extends Feature {

    public static final String TYPE = "http://schemas.xmlsoap.org/ws/2004/09/mex/";

    public static enum MetadataType {
        WSDL, SCHEMA, POLICY
    }

    /*
     * (non-Javadoc)
     * 
     * SPI
     */
    public interface Provider extends Feature.Provider {
        public InputSource spi_getMetadata(MetadataType type, String location) throws WSMexFeatureException;
    }
    
    public WSMexFeature(WSMexFeatureFactory factory, Feature.Provider provider) {
        super(factory, provider);
    }
    
    /**
     * Retreives metadata of indicated type from indicated location.
     * 
     * @param type
     * @param location
     * @return
     * @throws WSMexFeatureException
     */
    public InputSource getMetadata(MetadataType type, String location) throws WSMexFeatureException {
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            return ((Provider)provider).spi_getMetadata(type, location);
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

}
