/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.feature.ws.security;

import java.util.Collections;
import java.util.Set;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.component.engine.rt.FeatureModule.WSSpecification;

public abstract class WSSE implements WSSpecification {
    
    public static WSSE getInstance(String namespace) {
        if (namespace.equals(WSSE11.NAMESPACE_URI)) {
            return new WSSE11();
        } else {
            return null;
        }
    }

    public static final String NAMESPACE_PREFIX = "wsse";
    public static final String SECURITY = "Security";
    
    
    public static class WSSE11 extends WSSE {

        public static final String NAMESPACE_URI = "http://docs.oasis-open.org/wss/oasis-wss-wsecurity-secext-1.1.xsd";

        public static class Infoset implements WSSpecification.Infoset {
         
            public static final QName SECURITY_ELEM = new QName(NAMESPACE_URI, SECURITY);

        }

        public static class Policy implements WSSpecification.Policy {

            public static final String PVN = "http://docs.oasis-open.org/ws-sx/ws-securitypolicy/200702";

            public static class Infoset implements WSSpecification.Infoset {
            }

            public Infoset getInfoset() {
                return new Infoset() {
                };
            }
            
            /**
             * TODO: create API which manifests values defined via WSRMP 11 
             * @return
             */
            public boolean isFoo() {
                return false;
            }

        }

        public WSSE11() {
        }
        
        public WSSE11(Set<WSSpecification.Policy> alternatives) {
            this.alternatives = alternatives;
        }

        private Set<WSSpecification.Policy> alternatives;

        @Override
        public WSSpecification.Infoset getInfoSet() {
            return new Infoset() {
            };
        }

        @Override
        public String getNamespace() {
            return NAMESPACE_URI;
        }

        @Override
        @SuppressWarnings("unchecked")
        public Set<WSSpecification.Policy> getPolicy() {
            if (alternatives == null) {
                return Collections.EMPTY_SET;
            }
            return alternatives;
        }

    }

    public abstract Infoset getInfoSet();

    public abstract String getNamespace();

    public abstract Set<Policy> getPolicy();

}
