/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.feature.ws.security.soap;

import org.bluestemsoftware.specification.eoa.component.engine.rt.BindingModule;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointActionReference;
import org.bluestemsoftware.specification.eoa.component.policy.rt.ActionPolicy;
import org.bluestemsoftware.specification.eoa.component.policy.rt.UnsupportedPolicyException;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPBinding;
import org.bluestemsoftware.specification.eoa.ext.feature.FeatureException;
import org.bluestemsoftware.specification.eoa.ext.feature.FeatureFactory;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.mtom.soap.WSMTOMBindingModule;
import org.bluestemsoftware.specification.eoa.system.server.BindingFeature;

/**
 * A <code>ModulableFeature</code> which implements ws-security behavior within context of a
 * SOAP binding.
 * <p>
 * Note that version of ws-security specification implemented is provider specific as indicated
 * by the set of policy vocabularies understood by provider.
 */
public class WSSEFeature extends BindingFeature {

    /**
     * Binding type specific, ws-rm version agnostic URI which uniquely identifies this
     * feature.
     */
    public static final String TYPE = WSSEFeature.class.getName();

    public interface Provider extends BindingFeature.Provider {
        
        public WSMTOMBindingModule.Provider spi_createBindingModule(EndpointActionReference endpointActionReference, ActionPolicy privatePolicy, ActionPolicy publicPolicy) throws FeatureException, UnsupportedPolicyException;

        public WSMTOMBindingModule.Provider spi_createDefaultBindingModule() throws FeatureException;
        
    }

    public WSSEFeature(FeatureFactory factory, Provider provider) {
        super(factory, provider, SOAPBinding.BINDING_TYPE);
    }
    
    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.system.server.BindingFeature#createBindingModule(org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointActionReference, org.bluestemsoftware.specification.eoa.component.policy.rt.ActionPolicy, org.bluestemsoftware.specification.eoa.component.policy.rt.ActionPolicy)
     */
    @Override
    public BindingModule createBindingModule(EndpointActionReference endpointActionReference, ActionPolicy privatePolicy, ActionPolicy publicPolicy) throws FeatureException, UnsupportedPolicyException {
        if (provider == null) {
            throw new IllegalStateException("provider not set.");
        }
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            WSSEBindingModule.Provider p = ((Provider)provider).spi_createBindingModule(endpointActionReference, privatePolicy,
                    publicPolicy);
            return new WSSEBindingModule(p, this);
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

   /*
    * (non-Javadoc)
    * @see org.bluestemsoftware.specification.eoa.system.server.BindingFeature#createDefaultBindingModule()
    */
    public BindingModule createDefaultBindingModule() throws FeatureException {
        if (provider == null) {
            throw new IllegalStateException("provider not set.");
        }
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            WSSEBindingModule.Provider p =  ((Provider)provider).spi_createDefaultBindingModule();
            return new WSSEBindingModule(p, this);
        } finally {
            thread.setContextClassLoader(cl);
        }
    }
    
}
