/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.vm;

import java.util.Set;

import org.bluestemsoftware.specification.eoa.ext.feature.FeatureFactory;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.rt.TransportProtocolRT;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.vm.rt.VMTransportProtocolRT;
import org.bluestemsoftware.specification.eoa.system.server.TransportFeature;

/**
 * Creates <code>VMTransportModule</code> components which optimize message transmission by
 * sending/recieving messages within virtual memory, i.e. as opposed to transimitting over
 * 'wire'. The sending and receiving Endpoints MUST be defined on locally hosted Services.
 */
public class VMTransportFeature extends TransportFeature {

    public static final String TYPE = VMTransportFeature.class.getName();

    public static final String PVNS_10 = "http://bluestemsoftware.org/specification/eoa/ext/feature/ws/transport/vm/policy/1.0/";

    protected Set<String> policyVocabularyNamespaces;

    /*
     * (non-Javadoc)
     * 
     * SPI
     */
    public interface Provider extends TransportFeature.Provider {
    }

    public VMTransportFeature(FeatureFactory factory, Provider provider) {
        super(factory, provider);
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.system.server.TransportFeature#getTransportProtocol()
     */
    @Override
    public Class<? extends TransportProtocolRT> getTransportProtocol() {
        return VMTransportProtocolRT.class;
    }
    
}
