/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.vm.rt;

import org.bluestemsoftware.specification.eoa.component.binding.rt.BindingRT;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointReference;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction.Direction;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.rt.TransportProtocolRT;
import org.w3c.dom.Element;

/**
 * Interface implemented by instances of {@link BindingRT} which use the VM transport to
 * transmit messages.
 */
public interface VMTransportProtocolRT extends TransportProtocolRT {

    /**
     * Invoked by an instance of <code>VMTransportModule</code>. Receives a message
     * addressed to referenced endpoint.
     * <p>
     * Note the rules which define the behavior of an instance of
     * <code>VMTransportProtocolRT</code> are binding specific and MUST be defined by
     * concrete instances of this class.
     * 
     * @param endpointReference
     * @param direction
     *        indicates direction of message associated with action.
     * @param envelope
     *        Root element of a messaging protocol specific envelope, e.g. a SOAP envelope.
     */
    public void receiveAction(EndpointReference endpointReference, Direction direction, Element envelope);

}
