/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl11;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.xml.XMLConstants;
import javax.xml.namespace.QName;
import javax.xml.transform.Source;
import javax.xml.transform.dom.DOMSource;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.ComponentContext;
import org.bluestemsoftware.specification.eoa.component.ComponentDeployment;
import org.bluestemsoftware.specification.eoa.component.ComponentReader;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.RootComponent;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentName;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentType;
import org.bluestemsoftware.specification.eoa.component.application.Application;
import org.bluestemsoftware.specification.eoa.component.application.Role;
import org.bluestemsoftware.specification.eoa.component.binding.Binding;
import org.bluestemsoftware.specification.eoa.component.engine.Actor;
import org.bluestemsoftware.specification.eoa.component.engine.ApplicationReference;
import org.bluestemsoftware.specification.eoa.component.engine.Engine;
import org.bluestemsoftware.specification.eoa.component.intrface.Interface;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceFault;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceFaultReference;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceMessageReference;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceOperation;
import org.bluestemsoftware.specification.eoa.component.intrface.RequestMessageReference;
import org.bluestemsoftware.specification.eoa.component.intrface.ResponseMessageReference;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction.Direction;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceOperation.MEP;
import org.bluestemsoftware.specification.eoa.component.message.InterfaceMessage;
import org.bluestemsoftware.specification.eoa.component.message.MessagePart;
import org.bluestemsoftware.specification.eoa.component.policy.rt.ElementPolicy;
import org.bluestemsoftware.specification.eoa.component.policy.rt.PolicyExpression;
import org.bluestemsoftware.specification.eoa.component.service.BindingReference;
import org.bluestemsoftware.specification.eoa.component.service.Endpoint;
import org.bluestemsoftware.specification.eoa.component.service.InterfaceReference;
import org.bluestemsoftware.specification.eoa.component.service.Service;
import org.bluestemsoftware.specification.eoa.component.types.SchemaComponent;
import org.bluestemsoftware.specification.eoa.ext.binding.BindingFactory;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAP11HTTPProtocol;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPBinding;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPVersion;
import org.bluestemsoftware.specification.eoa.ext.idl.IDLDocument;
import org.bluestemsoftware.specification.eoa.ext.idl.IDLException;
import org.bluestemsoftware.specification.eoa.ext.idl.IDLFactory.BindingDefinitionReader.BindingStyle;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.WSDLDocument;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.WSDLException;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.WSDLFactory;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.WSDLLocation;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl11.PortTypeOperationElement.Style;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl11.ext.ActorElement;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl11.ext.ApplicationElement;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl11.ext.EngineElement;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl11.ext.RoleElement;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl11.ext.SOAP11AddressElement;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl11.ext.SOAPAddressElement;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl11.ext.SOAPBindingElement;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl11.ext.SOAPBodyElement;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl11.ext.SOAPFaultElement;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl11.ext.SOAPHeaderElement;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl11.ext.SOAPOperationElement;
import org.bluestemsoftware.specification.eoa.ext.policy.PolicyException;
import org.bluestemsoftware.specification.eoa.ext.policy.PolicyFactory;
import org.bluestemsoftware.specification.eoa.ext.schema.SchemaFactory;
import org.bluestemsoftware.specification.eoa.ext.schema.xs.xs10.XMLSchema;
import org.bluestemsoftware.specification.eoa.ext.service.HTTPEndpoint;
import org.bluestemsoftware.specification.eoa.system.SystemContext;
import org.bluestemsoftware.specification.eoa.system.System.Log;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * A WSDL factory used to read and write <code>RootComponent</code> objects. WSDL
 * consumed\generated complies with WSDL 1.1 as indicated by the type ref {@link #REF}.
 * <p>
 * Compliant implementations MUST insure that all wsdl 1.1 documents consumed/generated comply
 * with version 1.1 of the Basic Profile. Note that basic profile requires that all component
 * references be dereferenceable, i.e. a top-level component MUST NOT reference a component
 * defined within another deployment. Note that WSDL 2.0 has no such restriction.
 * 
 * @see http://www.ws-i.org/Profiles/BasicProfile-1.1.html
 * 
 */
public final class WSDL11Factory extends WSDLFactory {

    private static final Log log = SystemContext.getContext().getSystem().getLog(WSDL11Factory.class);

    public static final String WSDL_NS = WSDL11Document.TYPE;
    public static final String WSDL_NS_PREFIX = "wsdl";
    public static final String WSDL_SOAP11_NS = "http://schemas.xmlsoap.org/wsdl/soap/";
    public static final String WSDL_SOAP12_NS = "http://schemas.xmlsoap.org/wsdl/soap12/";
    public static final String WSDL_SOAP11_NS_PREFIX = "wsoap11";
    public static final String WSDL_SOAP12_NS_PREFIX = "wsoap12";
    public static final String VERSION_LABEL = "11";
    public static final String LINE_BREAK = System.getProperty("line.separator");

    private static TypesDefinitionReader typesDefinitionReader = new TypesDefinitionReader();
    private static PolicyDefinitionReader policyDefinitionReader = new PolicyDefinitionReader();
    private static EngineDefinitionReader engineDefinitionReader = new EngineDefinitionReader();
    private static ApplicationDefinitionReader applicationDefinitionReader = new ApplicationDefinitionReader();
    private static InterfaceDefinitionReader interfaceDefinitionReader = new InterfaceDefinitionReader();
    private static MessageDefinitionReader messageDefinitionReader = new MessageDefinitionReader();
    private static ServiceDefinitionReader serviceDefinitionReader = new ServiceDefinitionReader();
    private static BindingDefinitionReader bindingDefinitionReader = new BindingDefinitionReader();

    private static EngineDefinitionWriter engineDefinitionWriter = new EngineDefinitionWriter();
    private static ApplicationDefinitionWriter applicationDefinitionWriter = new ApplicationDefinitionWriter();
    private static InterfaceDefinitionWriter interfaceDefinitionWriter = new InterfaceDefinitionWriter();
    private static ServiceDefinitionWriter serviceDefinitionWriter = new ServiceDefinitionWriter();
    private static BindingDefinitionWriter bindingDefinitionWriter = new BindingDefinitionWriter();
    private static MessageDefinitionWriter messageDefinitionWriter = new MessageDefinitionWriter();

    /*
     * (non-Javadoc)
     * 
     * SPI
     */
    public interface Provider extends WSDLFactory.Provider {

        public WSDL11Document spi_createDocument(ComponentContext componentContext, Source source) throws WSDLException;

        public WSDL11Document spi_createDocument() throws WSDLException;

    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.ExtensionFactory#getExtensionType()
     */
    public final String getExtensionType() {
        return WSDL11Document.TYPE;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.idl.wsdl.WSDLFactory#getVersionLabel()
     */
    @Override
    public final String getVersionLabel() {
        return VERSION_LABEL;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReader#readComponents(org.bluestemsoftware.specification.eoa.component.ComponentContext,
     *      org.bluestemsoftware.specification.eoa.ext.idl.wsdl.WSDLDocument)
     */
    public void readComponents(ComponentContext componentContext, IDLDocument idlDocument) throws DeploymentException {

        if (factoryContext == null) {
            throw new IllegalArgumentException("deploymentContext null");
        }
        if (idlDocument == null) {
            throw new IllegalArgumentException("source null");
        }

        if (!(idlDocument instanceof WSDL11Document)) {
            throw new DeploymentException("Expected wsdlDocument of type " + WSDL11Document.TYPE);
        }

        log.trace("readComponents begin");

        readTopLevelComponents(componentContext, (WSDL11Document)idlDocument);

        log.trace("readComponents end");

    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.idl.wsdl.WSDLFactory#readWSDL(org.bluestemsoftware.specification.eoa.component.ComponentContext,
     *      javax.xml.transform.Source)
     */
    @Override
    public WSDLDocument readIDL(ComponentContext componentContext, Source source) throws WSDLException {
        if (factoryContext == null) {
            throw new IllegalStateException("Factory has no context.");
        }
        if (componentContext == null) {
            throw new IllegalArgumentException("componentContext null");
        }
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factoryContext.getClassLoader());
            return ((Provider)provider).spi_createDocument(componentContext, source);
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.idl.wsdl.WSDLFactory#writeWSDL(org.bluestemsoftware.specification.eoa.ext.idl.wsdl.WSDLLocation)
     */
    public WSDLDocument writeWSDL(WSDLLocation wsdlLocation) throws IDLException {

        if (factoryContext == null) {
            throw new IllegalStateException("Factory has no context.");
        }
        if (wsdlLocation == null) {
            throw new IllegalArgumentException("wsdlLocation null");
        }

        WSDL11Document wsdlDocument = null;
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factoryContext.getClassLoader());
            wsdlDocument = ((Provider)provider).spi_createDocument();
        } finally {
            thread.setContextClassLoader(cl);
        }

        wsdlDocument.setDocumentURI(wsdlLocation.toURI().toString());
        DefinitionsElement def = wsdlDocument.createDefinitionsElement();

        Set<FragmentIdentifier> fids = wsdlLocation.getFragmentIdentifiers();

        if (fids.size() == 0) {
            return wsdlDocument;
        }

        String tns = fids.iterator().next().getRootComponentNamespace();

        def.setTargetNamespace(tns);
        wsdlDocument.setDefinitionsElement(def);

        for (FragmentIdentifier fid : fids) {
            RootComponent component = SystemContext.getContext().getSystem().getComponent(fid);
            if (component == null) {
                throw new WSDLException("Component " + fid + " is undefined.");
            }
            if (component.getComponentType() == ComponentType.APPLICATION) {
                applicationDefinitionWriter.writeDefinition(def, (Application)component);
            }
            if (component.getComponentType() == ComponentType.BINDING) {
                bindingDefinitionWriter.writeDefinition(def, (Binding)component);
            }
            if (component.getComponentType() == ComponentType.ENGINE) {
                engineDefinitionWriter.writeDefinition(def, (Engine)component);
            }
            if (component.getComponentType() == ComponentType.INTERFACE) {
                interfaceDefinitionWriter.writeDefinition(def, (Interface)component);
            }
            if (component.getComponentType() == ComponentType.SERVICE) {
                serviceDefinitionWriter.writeDefinition(def, (Service)component);
            }
            if (component.getComponentType() == ComponentType.MESSAGE) {
                messageDefinitionWriter.writeDefinition(def, (InterfaceMessage)component);
            }
        }

        // iterate over imports and set location attribute using wsdlLocation
        // instance cached on import elements by writers

        for (ImportElement importElement : def.getImportElements().values()) {
            WSDLLocation location = ((ImportElement)importElement).getWsdlLocation();
            importElement.setLocation(location.toString());
        }

        return wsdlDocument;

    }

    /** ***************************** Component Readers ********************************** * */

    private static class ApplicationDefinitionReader {

        public Application readDefinition(DefinitionsElement def, ComponentContext deploymentContext, ComponentName componentName) throws DeploymentException {

            ComponentDeployment componentDeployment = deploymentContext.getDeployment();

            log.trace("readDefinition begin");

            ApplicationElement applicationElement = null;

            if (def.getTargetNamespace().equals(componentName.getNamespaceURI())) {

                // attempt to retrieve from definition
                applicationElement = def.getApplicationElements().get(componentName.getLocalPart());

            } else {

                // check each import for component definition
                for (ImportElement importElement : def.getImportElements().values()) {
                    if (importElement.getNamespace().equals(componentName.getNamespaceURI())) {
                        DefinitionsElement temp;
                        try {
                            temp = importElement.getDocument().getDefinitionsElement();
                        } catch (WSDLException we) {
                            throw new DeploymentException(we.getMessage());
                        }
                        applicationElement = temp.getApplicationElements().get(componentName.getLocalPart());
                        if (applicationElement != null) {
                            break;
                        }
                    }
                }

            }

            // if not found, this is an error. unlike wsdl 20, wsdl 11 requires a resolvable
            // location attrib on import elements, i.e. def cannot ref components in other
            // deployments. also, basic profile validation requires that all components be
            // dereferenced
            if (applicationElement == null) {
                throw new DeploymentException("Component "
                        + componentName
                        + " not found/dereferenceable within/from WSDL 1.1 definition");
            }

            // if we've already parsed this component, then just retrieve the existing
            // definition from deployment

            if (applicationElement.isRead()) {
                FragmentIdentifier fid = FragmentIdentifier.valueOf(ComponentType.APPLICATION, componentName);
                return (Application)componentDeployment.getComponent(fid);
            } else {
                applicationElement.setIsRead(true);
            }

            Map<String, Role> roles = new HashMap<String, Role>();

            for (RoleElement roleElement : applicationElement.getRoleElements().values()) {
                String roleName = roleElement.getName();
                ComponentName interfaceName = roleElement.getInterface();
                roles.put(roleName, new Role(roleName, interfaceName));
            }

            Application application = new Application(componentName, roles);
            componentDeployment.addComponent(application);

            for (Role role : application.getRoles()) {
                interfaceDefinitionReader
                        .readDefinition(def, deploymentContext, role.getReferencedComponentName());
            }

            log.trace("readDefinition end");
            return application;
        }

    }

    private static class BindingDefinitionReader {

        public Binding readDefinition(DefinitionsElement def, ComponentContext deploymentContext, ComponentName componentName) throws DeploymentException {

            ComponentDeployment componentDeployment = deploymentContext.getDeployment();

            log.trace("readDefinition begin");

            BindingElement bindingElement = null;

            if (def.getTargetNamespace().equals(componentName.getNamespaceURI())) {

                // attempt to retrieve from definition
                bindingElement = def.getBindingElements().get(componentName.getLocalPart());

            } else {

                // check each import for component definition
                for (ImportElement importElement : def.getImportElements().values()) {
                    if (importElement.getNamespace().equals(componentName.getNamespaceURI())) {
                        DefinitionsElement temp;
                        try {
                            temp = importElement.getDocument().getDefinitionsElement();
                        } catch (WSDLException we) {
                            throw new DeploymentException(we.getMessage());
                        }
                        bindingElement = temp.getBindingElements().get(componentName.getLocalPart());
                        if (bindingElement != null) {
                            break;
                        }
                    }
                }

            }

            // if not found, this is an error. unlike wsdl 20, wsdl 11 requires a resolvable
            // location attrib on import elements, i.e. def cannot ref components in other
            // deployments. also, basic profile validation requires that all components be
            // dereferenced
            if (bindingElement == null) {
                throw new DeploymentException("Component "
                        + componentName
                        + " not found/dereferenceable within/from WSDL 1.1 definition");
            }

            // if we've already parsed this component, then just retrieve the existing
            // definition from deployment

            if (bindingElement.isRead()) {
                FragmentIdentifier fid = FragmentIdentifier.valueOf(ComponentType.BINDING, componentName);
                return (Binding)componentDeployment.getComponent(fid);
            } else {
                bindingElement.setIsRead(true);
            }

            // validate defintion against basic profile
            ValidationInfo validationInfo = BasicProfile.validateBinding(def, componentName);
            ValidationMessage validationMessage = validationInfo.getValidationMessage();
            if (validationMessage.length() > 0) {
                String error = LINE_BREAK
                        + "Error reading definition for component "
                        + componentName
                        + ". WSDL 1.1 definition failed validation with the following errors:"
                        + LINE_BREAK
                        + validationMessage.toString()
                        + LINE_BREAK;
                throw new DeploymentException(error);
            }

            // attempt to dereference the interface. run request through
            // reader rather than helper in case interface is defined on an
            // imported def

            ComponentName interfaceName = bindingElement.getType();
            interfaceDefinitionReader.readDefinition(def, deploymentContext, interfaceName);

            // using binding type ref, retreive a factory which understands binding
            // type and can parse binding element into a binding component. note
            // that the implementation is arbitrary, i.e. impls for the same binding
            // type are extension factory providers for the same binding factory
            // instance

            BindingFactory bindingFactory = null;
            for (BindingFactory tmp : SystemContext.getContext().getSystem().getBindingFactories()) {
                if(tmp.getBindingType().equals(SOAPBinding.BINDING_TYPE)) {
                    bindingFactory = tmp;
                    break;
                }
            }
            
            if (bindingFactory == null) {
                throw new DeploymentException("Error reading definition for Binding "
                        + componentName
                        + ". BindingFactory of type '"
                        + SOAPBinding.BINDING_TYPE
                        + "' is undefined.");
            }

            try {
                bindingFactory.readBinding(deploymentContext, def.getTargetNamespace(), bindingElement,
                        validationInfo.getBindingStyle());
            } catch (DeploymentException de) {
                throw new DeploymentException("Error reading definition for Binding " + componentName + ". " + de);
            }

            log.trace("readBinding end");
            return (Binding)componentDeployment.getComponent(ComponentType.BINDING, componentName);

        }

    }

    private static class EngineDefinitionReader {

        public Engine readDefinition(DefinitionsElement def, ComponentContext deploymentContext, ComponentName componentName) throws DeploymentException {

            ComponentDeployment componentDeployment = deploymentContext.getDeployment();

            log.trace("readDefinition begin");

            EngineElement engineElement = null;

            if (def.getTargetNamespace().equals(componentName.getNamespaceURI())) {

                // attempt to retrieve from definition
                engineElement = def.getEngineElements().get(componentName.getLocalPart());

            } else {

                // check each import for component definition
                for (ImportElement importElement : def.getImportElements().values()) {
                    if (importElement.getNamespace().equals(componentName.getNamespaceURI())) {
                        DefinitionsElement temp;
                        try {
                            temp = importElement.getDocument().getDefinitionsElement();
                        } catch (WSDLException we) {
                            throw new DeploymentException(we.getMessage());
                        }
                        engineElement = temp.getEngineElements().get(componentName.getLocalPart());
                        if (engineElement != null) {
                            break;
                        }
                    }
                }

            }

            // if not found, this is an error. unlike wsdl 20, wsdl 11 requires a resolvable
            // location attrib on import elements, i.e. def cannot ref components in other
            // deployments. also, basic profile validation requires that all components be
            // dereferenced
            if (engineElement == null) {
                throw new DeploymentException("Component "
                        + componentName
                        + " not found/dereferenceable within/from WSDL 1.1 definition");
            }

            // if we've already parsed this component, then just retrieve the existing
            // definition from deployment

            if (engineElement.isRead()) {
                FragmentIdentifier fid = FragmentIdentifier.valueOf(ComponentType.ENGINE, componentName);
                return (Engine)componentDeployment.getComponent(fid);
            } else {
                engineElement.setIsRead(true);
            }

            // now parse engineElement into an engine component
            Set<Actor> actors = new HashSet<Actor>();
            for (ActorElement actorElement : engineElement.getActorElements().values()) {
                ComponentName serviceName = actorElement.getService();
                String role = actorElement.getRole();
                actors.add(new Actor(role, serviceName));
            }

            Engine engine = new Engine(componentName, new ApplicationReference(engineElement.getApplication()),
                    actors);
            componentDeployment.addComponent(engine);

            applicationDefinitionReader.readDefinition(def, deploymentContext, engine.getApplicationReference()
                    .getReferencedComponentName());

            for (Actor actor : actors) {
                ComponentName serviceName = actor.getReferencedComponentName();
                serviceDefinitionReader.readDefinition(def, deploymentContext, serviceName);
            }

            log.trace("readEngine end");
            return engine;

        }

    }

    private final static String WSA_METADATA_NS = "http://www.w3.org/2007/05/addressing/metadata";
    private final static String ATT_WSA_ACTION = "Action";
    private static final String SOAP_HTTP_TRANSPORT = "http://schemas.xmlsoap.org/soap/http";
    private static final String ENCODED_USE = "encoded";
    private static final String RPC_BINDING_STYLE = "rpc";

    private static class InterfaceDefinitionReader {

        public Interface readDefinition(DefinitionsElement def, ComponentContext deploymentContext, ComponentName componentName) throws DeploymentException {

            ComponentDeployment componentDeployment = deploymentContext.getDeployment();

            log.trace("readDefinition begin");

            PortTypeElement portTypeElement = null;

            if (def.getTargetNamespace().equals(componentName.getNamespaceURI())) {

                // attempt to retrieve from definition
                portTypeElement = def.getPortTypeElements().get(componentName.getLocalPart());

            } else {

                // check each import for component definition
                for (ImportElement importElement : def.getImportElements().values()) {
                    if (importElement.getNamespace().equals(componentName.getNamespaceURI())) {
                        DefinitionsElement temp;
                        try {
                            temp = importElement.getDocument().getDefinitionsElement();
                        } catch (WSDLException we) {
                            throw new DeploymentException(we.getMessage());
                        }
                        portTypeElement = temp.getPortTypeElements().get(componentName.getLocalPart());
                        if (portTypeElement != null) {
                            break;
                        }
                    }
                }

            }

            // if not found, this is an error. unlike wsdl 20, wsdl 11 requires a resolvable
            // location attrib on import elements, i.e. def cannot ref components in other
            // deployments. also, basic profile validation requires that all components be
            // dereferenced
            if (portTypeElement == null) {
                throw new DeploymentException("Component "
                        + componentName
                        + " not found/dereferenceable within/from WSDL 1.1 definition");
            }

            // if we've already parsed this component, then just retrieve the existing
            // definition from deployment

            if (portTypeElement.isRead()) {
                FragmentIdentifier fid = FragmentIdentifier.valueOf(ComponentType.INTERFACE, componentName);
                return (Interface)componentDeployment.getComponent(fid);
            } else {
                portTypeElement.setIsRead(true);
            }

            // validate defintion against basic profile
            ValidationInfo validationInfo = BasicProfile.validatePortType(def, componentName);
            ValidationMessage validationMessage = validationInfo.getValidationMessage();
            if (validationMessage.length() > 0) {
                String error = LINE_BREAK
                        + "Error reading definition for component "
                        + componentName
                        + ". WSDL 1.1 definition failed validation with the following errors:"
                        + LINE_BREAK
                        + validationMessage.toString()
                        + LINE_BREAK;
                throw new DeploymentException(error);
            }

            // now parse portTypeElement into an interface component

            Map<QName, InterfaceFault> interfaceFaults = new HashMap<QName, InterfaceFault>();
            Map<QName, InterfaceOperation> interfaceOperations = new HashMap<QName, InterfaceOperation>();
            Map<QName, Map<QName, InterfaceFaultReference>> operationFaults = new HashMap<QName, Map<QName, InterfaceFaultReference>>();

            // we must iterate over operations and create each before creating
            // interface. this is because we must process all faults first.
            // our component model accommodates both wsdl 2.x and wsdl 1.1.
            // consequently faults are defined at interface level rather than
            // at operation granularity, i.e. we will not support faults with
            // same name and different message types within a single interface

            for (PortTypeOperationElement portTypeOperation : portTypeElement.getOperationElements().values()) {

                QName operationName = new QName(componentName.getNamespaceURI(), portTypeOperation.getName());

                log.debug("processing interface operation '" + operationName.getLocalPart() + "'");

                // derive message exchange pattern used by this operation
                MEP mep = null;
                if (portTypeOperation.getOutputElement() != null) {
                    mep = MEP.IN_OUT;
                } else {
                    mep = MEP.IN_ONLY;
                }

                // process the referenced input message. run request through def
                // reader rather than helper in case message is defined on an
                // imported def

                PortTypeOperationInputElement inputElement = portTypeOperation.getInputElement();
                String inputName = inputElement.getName() == "" ? null : inputElement.getName();
                ComponentName inputMessageName = inputElement.getMessage();
                log.debug("processing wsdl input message " + inputMessageName);
                messageDefinitionReader.readDefinition(def, deploymentContext, inputMessageName);

                // parse element policy attached to message reference element
                ElementPolicy elementPolicy = PolicyUtil.getAttachedPolicy(deploymentContext, new DOMSource(
                        inputElement, def.getTargetNamespace()));

                // check for an explicit wsa action property. if one doesn't
                // exist, we'll use the default action pattern for wsdl 1.1
                // as specified in wsa spec

                String action = inputElement.getAttributeNS(WSA_METADATA_NS, ATT_WSA_ACTION);

                if (action.equals("")) {
                    action = createDefaultMessageReferenceAction(componentName, operationName.getLocalPart(), mep,
                            inputName, Direction.IN);
                }

                // create messageReference object. note that contentModel is undefined,
                // i.e. no assertion can be made regarding how message payload is modeled
                // when interface def is imported via wsdl 1.1 without a wsdl 1.1 binding.
                // interface cannot be described via wsdl 2.0

                InterfaceMessageReference inputMessageRef = new RequestMessageReference(inputMessageName, action,
                        Direction.IN, null, elementPolicy);

                // process the optional output message
                PortTypeOperationOutputElement outputElement = portTypeOperation.getOutputElement();
                String outputName = null;
                ComponentName outputMessageName = null;
                InterfaceMessageReference outputMessageRef = null;

                if (outputElement != null) {

                    // process the referenced output message. run request through
                    // def reader rather than helper in case message is defined on
                    // an imported def

                    outputName = outputElement.getName() == "" ? null : outputElement.getName();
                    outputMessageName = outputElement.getMessage();
                    log.debug("processing wsdl output message " + outputMessageName);
                    messageDefinitionReader.readDefinition(def, deploymentContext, outputMessageName);

                    // parse element policy attached to message reference element
                    elementPolicy = PolicyUtil.getAttachedPolicy(deploymentContext, new DOMSource(outputElement,
                            def.getTargetNamespace()));

                    // check for an explicit wsa action property. if one doesn't
                    // exist, we'll use the default action pattern for wsdl 1.1
                    // as specified in wsa spec

                    action = outputElement.getAttributeNS(WSA_METADATA_NS, ATT_WSA_ACTION);

                    if (action.equals("")) {
                        action = createDefaultMessageReferenceAction(componentName, operationName.getLocalPart(),
                                mep, outputName, Direction.OUT);
                    }

                    // create messageReference object. note that contentModel is undefined,
                    // i.e. no assertion can be made regarding how message payload is modeled
                    // when interface def is imported via wsdl 1.1 without a wsdl 1.1 binding.
                    // interface cannot be described via wsdl 2.0

                    outputMessageRef = new ResponseMessageReference(outputMessageName, action, Direction.OUT,
                            null, elementPolicy);

                }

                // now process all optional wsdl:faults. no need to validate anything, as we
                // have already validated and can assume a ws-i compliant definition

                Map<QName, InterfaceFaultReference> faultRefs = new HashMap<QName, InterfaceFaultReference>();

                for (PortTypeOperationFaultElement faultElement : portTypeOperation.getFaultElements().values()) {

                    ComponentName faultMessageName = faultElement.getMessage();
                    messageDefinitionReader.readDefinition(def, deploymentContext, faultMessageName);

                    // because within wsdl 1.1 interfaces cannot be inherited, we build
                    // qualified name of fault using namespace of wsdl:portType within
                    // which fault is defined

                    String faultNS = componentName.getNamespaceURI();
                    QName faultName = new QName(faultNS, faultElement.getName());
                    log.debug("processing fault " + faultName);

                    if (interfaceFaults.containsKey(faultName.getLocalPart())) {
                        InterfaceFault interfaceFault = interfaceFaults.get(faultName.getLocalPart());
                        if (!interfaceFault.getReferencedComponentName().equals(faultMessageName)) {
                            throw new DeploymentException("Error reading portType "
                                    + componentName
                                    + ". Operation '"
                                    + operationName.getLocalPart()
                                    + "' is invalid. At least one other operation"
                                    + " throws same fault '"
                                    + faultName.getLocalPart()
                                    + "' but with a different message type. This"
                                    + " configuration is not supported.");
                        }
                    } else {

                        // there is no attachment point for an interface fault within
                        // wsdl 1.1
                        elementPolicy = null;

                        InterfaceFault interfaceFault = new InterfaceFault(faultName, faultMessageName,
                                elementPolicy);
                        interfaceFaults.put(faultName, interfaceFault);

                        // check for an explicit wsa action property. if one doesn't
                        // exist, we'll use the default action pattern for wsdl 1.1
                        // as specified in wsa spec

                        action = faultElement.getAttributeNS(WSA_METADATA_NS, ATT_WSA_ACTION);

                        if (action.equals("")) {
                            action = createDefaultFaultReferenceAction(componentName,
                                    operationName.getLocalPart(), faultName.getLocalPart());
                        }

                        // parse element policy attached to message reference element
                        elementPolicy = PolicyUtil.getAttachedPolicy(deploymentContext, new DOMSource(
                                faultElement, def.getTargetNamespace()));

                        // create faultReference object using factory and add to collection
                        InterfaceFaultReference faultRef = new InterfaceFaultReference(faultName, action,
                                Direction.OUT, elementPolicy);
                        faultRefs.put(faultName, faultRef);

                    }

                }

                operationFaults.put(operationName, faultRefs);

                // parse element policy attached to message reference element
                elementPolicy = PolicyUtil.getAttachedPolicy(deploymentContext, new DOMSource(portTypeOperation,
                        def.getTargetNamespace()));

                // operation style attribute is undefined for wsdl 1.1
                Set<String> styles = new HashSet<String>();

                // create and add operation using our factory methods
                InterfaceOperation interfaceOperation = new InterfaceOperation(operationName, mep, styles, false,
                        inputMessageRef, outputMessageRef, operationFaults.get(operationName), elementPolicy);
                interfaceOperations.put(operationName, interfaceOperation);

            }

            // parse element policy attached to message reference element
            ElementPolicy elementPolicy = PolicyUtil.getAttachedPolicy(deploymentContext, new DOMSource(
                    portTypeElement, def.getTargetNamespace()));

            // interface inheritance is not possible with wsdl 1.1, so
            // pass an empty set
            Set<QName> extendedInterfaces = new HashSet<QName>();

            // default style is undefined for wsdl 1.1.
            Set<String> defaultStyles = new HashSet<String>();
            Interface intrface = new Interface(componentName, extendedInterfaces, interfaceOperations,
                    interfaceFaults, defaultStyles, elementPolicy);
            componentDeployment.addComponent(intrface);

            log.trace("readPortType end");
            return intrface;

        }

        /*
         * Generates an action property for an interface message reference using Default Action
         * Pattern.
         * 
         * with explicit input/output message 'name' attribute
         * 
         * http://myserver.mycompany.com/component/interface/1.0/myInterface/myInputMessageName
         * http://myserver.mycompany.com/component/interface/1.0/myInterface/myOutputMessageName
         * 
         * without explicit input/output message 'name' attribute
         * 
         * http://myserver.mycompany.com/component/interface/1.0/myInterface/myOperationRequest
         * http://myserver.mycompany.com/component/interface/1.0/myInterface/myOperationResponse
         */

        private String createDefaultMessageReferenceAction(QName interfaceName, String operationName, MEP mep, String messageName, Direction direction) {
            String tns = interfaceName.getNamespaceURI();
            if (!tns.endsWith("/")) {
                tns = tns + "/";
            }
            StringBuilder sb = new StringBuilder();
            sb.append(tns);
            sb.append(interfaceName.getLocalPart());
            sb.append("/");
            if (messageName == null) {
                sb.append(operationName);
                if (direction == Direction.IN) {
                    sb.append("Request");
                } else {
                    sb.append("Response");
                }
            } else {
                sb.append(messageName);
            }
            return sb.toString();
        }

        /*
         * Generates an action property for an interface fault reference using Default Action
         * Pattern
         * 
         * http://myserver.mycompany.com/component/interface/1.0/myInterface/myOperation/Fault/myFault
         */
        private String createDefaultFaultReferenceAction(QName interfaceName, String operationName, String faultName) {
            String tns = interfaceName.getNamespaceURI();
            if (!tns.endsWith("/")) {
                tns = tns + "/";
            }
            StringBuilder sb = new StringBuilder();
            sb.append(tns);
            sb.append(interfaceName.getLocalPart());
            sb.append("/");
            sb.append(operationName);
            sb.append("/");
            sb.append("Fault");
            sb.append("/");
            sb.append(faultName);
            return sb.toString();
        }

    }

    private static class MessageDefinitionReader {

        public InterfaceMessage readDefinition(DefinitionsElement def, ComponentContext deploymentContext, ComponentName componentName) throws DeploymentException {

            ComponentDeployment componentDeployment = deploymentContext.getDeployment();

            log.trace("readDefinition begin");

            MessageElement messageElement = null;

            if (def.getTargetNamespace().equals(componentName.getNamespaceURI())) {

                // attempt to retrieve from definition
                messageElement = def.getMessageElements().get(componentName.getLocalPart());

            } else {

                // check each import for component definition
                for (ImportElement importElement : def.getImportElements().values()) {
                    if (importElement.getNamespace().equals(componentName.getNamespaceURI())) {
                        DefinitionsElement temp;
                        try {
                            temp = importElement.getDocument().getDefinitionsElement();
                        } catch (WSDLException we) {
                            throw new DeploymentException(we.getMessage());
                        }
                        messageElement = temp.getMessageElements().get(componentName.getLocalPart());
                        if (messageElement != null) {
                            break;
                        }
                    }
                }

            }

            // if not found, this is an error. unlike wsdl 20, wsdl 11 requires a resolvable
            // location attrib on import elements, i.e. def cannot ref components in other
            // deployments. also, basic profile validation requires that all components be
            // dereferenced
            if (messageElement == null) {
                throw new DeploymentException("Component "
                        + componentName
                        + " not found/dereferenceable within/from WSDL 1.1 definition");
            }

            // if we've already parsed this component, then just retrieve the existing
            // definition from deployment

            if (messageElement.isRead()) {
                FragmentIdentifier fid = FragmentIdentifier.valueOf(ComponentType.MESSAGE, componentName);
                return (InterfaceMessage)componentDeployment.getComponent(fid);
            } else {
                messageElement.setIsRead(true);
            }

            // validate defintion against basic profile
            ValidationInfo validationInfo = BasicProfile.validateMessage(def, componentName);
            ValidationMessage validationMessage = validationInfo.getValidationMessage();
            if (validationMessage.length() > 0) {
                String error = LINE_BREAK
                        + "Error reading definition for component "
                        + componentName
                        + ". WSDL 1.1 definition failed validation with the following errors:"
                        + LINE_BREAK
                        + validationMessage.toString()
                        + LINE_BREAK;
                throw new DeploymentException(error);
            }

            // parse messageElement into an message component

            ElementPolicy elementPolicy = PolicyUtil.getAttachedPolicy(deploymentContext, new DOMSource(
                    messageElement, def.getTargetNamespace()));

            Map<String, MessagePart> messageParts = new HashMap<String, MessagePart>();

            for (MessagePartElement partElement : messageElement.getPartElements().values()) {

                SchemaComponent.Type type = null;
                QName partType = null;
                if (partElement.getElement() != null) {
                    type = SchemaComponent.Type.ELEMENT_DECLARATION;
                    partType = partElement.getElement();
                } else {
                    type = SchemaComponent.Type.TYPE_DEFINITION;
                    partType = partElement.getType();
                }

                // note that reference to schema components are established at runtime
                // which allows for possibility that schema may have been deployed
                // within a separate deployment and referenced via an import with no
                // location attribute

                MessagePart abstractPart = new MessagePart(partElement.getName(), partType.getNamespaceURI(),
                        type, partType.getLocalPart());
                messageParts.put(abstractPart.getName(), abstractPart);

            }

            InterfaceMessage interfaceMessage = new InterfaceMessage(componentName, elementPolicy, messageParts);
            componentDeployment.addComponent(interfaceMessage);

            log.trace("readMessage end");
            return interfaceMessage;

        }

    }

    private static class PolicyDefinitionReader {

        public void readPolicy(DefinitionsElement def, ComponentContext deploymentContext) throws DeploymentException {
            log.trace("readPolicy begin");

            for (PolicyFactory policyFactory : SystemContext.getContext().getSystem().getPolicyFactories()) {
                policyFactory.readComponents(deploymentContext, (WSDL11Document)def.getOwnerDocument());
            }

            log.trace("readPolicy end");
        }

    }

    private static class ServiceDefinitionReader {

        public Service readDefinition(DefinitionsElement def, ComponentContext deploymentContext, ComponentName componentName) throws DeploymentException {

            ComponentDeployment componentDeployment = deploymentContext.getDeployment();

            log.trace("readDefinition begin");

            ServiceElement serviceElement = null;

            if (def.getTargetNamespace().equals(componentName.getNamespaceURI())) {

                // attempt to retrieve from definition
                serviceElement = def.getServiceElements().get(componentName.getLocalPart());

            } else {

                // check each import for component definition
                for (ImportElement importElement : def.getImportElements().values()) {
                    if (importElement.getNamespace().equals(componentName.getNamespaceURI())) {
                        DefinitionsElement temp;
                        try {
                            temp = importElement.getDocument().getDefinitionsElement();
                        } catch (WSDLException we) {
                            throw new DeploymentException(we.getMessage());
                        }
                        serviceElement = temp.getServiceElements().get(componentName.getLocalPart());
                        if (serviceElement != null) {
                            break;
                        }
                    }
                }

            }

            // if not found, this is an error. unlike wsdl 20, wsdl 11 requires a resolvable
            // location attrib on import elements, i.e. def cannot ref components in other
            // deployments. also, basic profile validation requires that all components be
            // dereferenced
            if (serviceElement == null) {
                throw new DeploymentException("Component "
                        + componentName
                        + " not found/dereferenceable within/from WSDL 1.1 definition");
            }

            // if we've already parsed this component, then just retrieve the existing
            // definition from deployment

            if (serviceElement.isRead()) {
                FragmentIdentifier fid = FragmentIdentifier.valueOf(ComponentType.SERVICE, componentName);
                return (Service)componentDeployment.getComponent(fid);
            } else {
                serviceElement.setIsRead(true);
            }

            // validate defintion against basic profile
            ValidationInfo validationInfo = BasicProfile.validateService(def, componentName);
            ValidationMessage validationMessage = validationInfo.getValidationMessage();
            if (validationMessage.length() > 0) {
                String error = LINE_BREAK
                        + "Error reading definition for component "
                        + componentName
                        + ". WSDL 1.1 definition failed validation with the following errors:"
                        + LINE_BREAK
                        + validationMessage.toString()
                        + LINE_BREAK;
                throw new DeploymentException(error);
            }

            // parse serviceElement into an service component
            Map<String, Endpoint> endpoints = new HashMap<String, Endpoint>();
            for (ServicePortElement portElement : serviceElement.getPortElements().values()) {
                Endpoint endpoint = readPort(def, deploymentContext, componentName, portElement);
                endpoints.put(endpoint.getName(), endpoint);
            }

            // we need to derive interface implemented by service by retrieving from its
            // bindings. basic profile insures that they all match
            BindingReference br = endpoints.values().iterator().next().getBindingReference();
            QName bindingName = br.getReferencedComponentName();
            FragmentIdentifier reference = FragmentIdentifier.valueOf(ComponentType.BINDING, bindingName);
            Binding binding = (Binding)componentDeployment.getComponent(reference);
            ComponentName interfaceName = binding.getInterfaceReference().getReferencedComponentName();

            // pares element policy attached to message reference element
            ElementPolicy elementPolicy = PolicyUtil.getAttachedPolicy(deploymentContext, new DOMSource(
                    serviceElement, def.getTargetNamespace()));

            // create service component. profile insures that at least one
            // port is defined. service def is loaded by caller
            Service service = new Service(componentName, new InterfaceReference(interfaceName), elementPolicy,
                    endpoints);
            componentDeployment.addComponent(service);

            log.trace("readService end");
            return service;

        }

        private Endpoint readPort(DefinitionsElement def, ComponentContext deploymentContext, ComponentName serviceName, ServicePortElement portElement) throws DeploymentException {
            log.trace("readPort begin");

            SOAPAddressElement soapAddress = portElement.getSOAPAddressElement();

            if (soapAddress == null) {
                throw new DeploymentException("Error reading endpoint on service "
                        + serviceName
                        + ". Failed to find valid SOAPAddress within port '"
                        + portElement.getName()
                        + "'.");
            }
            String address = soapAddress.getLocation();
            if (address == null) {
                throw new DeploymentException("Error reading endpoint on service "
                        + serviceName
                        + ". LocationURI specified within SOAPAddress for port '"
                        + portElement.getName()
                        + "' is null or empty string");
            }

            // because wsdl 2.0 decorates endpoint with transport specific attribs
            // which should have been done with ws-policy, we are forced to create
            // an HTTPEndpoint extension if address uses http scheme

            boolean isHTTP = false;
            try {
                URI temp = new URI(address);
                if (temp.getScheme().equalsIgnoreCase("http")) {
                    isHTTP = true;
                }
            } catch (URISyntaxException ue) {
                throw new DeploymentException("Error reading endpoint on service "
                        + serviceName
                        + ". Failed to parse address. "
                        + ue.getMessage());
            }

            // read binding associated with the port we're reading. binding
            // helper will load definition into component model

            ComponentName bindingName = portElement.getBinding();
            bindingDefinitionReader.readDefinition(def, deploymentContext, bindingName);

            // pares element policy attached to endpoint element
            ElementPolicy elementPolicy = PolicyUtil.getAttachedPolicy(deploymentContext, new DOMSource(
                    portElement, def.getTargetNamespace()));

            // create the endpoint
            Endpoint endpoint = null;
            if (isHTTP) {
                endpoint = new HTTPEndpoint(portElement.getName(), new BindingReference(bindingName), address,
                        null, null, elementPolicy);
            } else {
                endpoint = new Endpoint(portElement.getName(), new BindingReference(bindingName), address,
                        elementPolicy);
            }
            
            log.trace("readPort end");
            return endpoint;
        }

    }

    private static class TypesDefinitionReader {

        public void readTypes(DefinitionsElement desc, ComponentContext deploymentContext) throws DeploymentException {

            log.trace("readTypes begin");

            // iterate over all schema factories defined on system and allow them to
            // read schema component definitions which may be defined within types el

            Set<SchemaFactory> schemaFactories = SystemContext.getContext().getSystem().getSchemaFactories();
            if (schemaFactories.size() == 0) {
                log.warn("no schema factories defined");
            }

            for (SchemaFactory schemaFactory : schemaFactories) {
                ((ComponentReader)schemaFactory).readComponents(deploymentContext, (WSDL11Document)desc
                        .getOwnerDocument());
            }

            log.trace("readTypes end");

        }

    }

    /** ***************************** Component Writers ********************************** * */

    private static class ApplicationDefinitionWriter {

        public void writeDefinition(DefinitionsElement def, Application application) throws WSDLException {

            if (def.getApplicationElements().get(application.getName()) != null) {
                return;
            }

            log.trace("writeDefinition begin");

            // first we add the application element to the definition

            WSDL11Document wsdlDocument = (WSDL11Document)def.getOwnerDocument();
            ApplicationElement applicationElement = wsdlDocument.createApplicationElement();
            def.addApplicationElement(applicationElement);
            applicationElement.setName(application.getName().getLocalPart());

            for (Role role : application.getRoles()) {
                RoleElement roleElement = wsdlDocument.createRoleElement();
                roleElement.setName(role.getName());
                roleElement.setInterface(role.getReferencedComponentName());
                applicationElement.addRoleElement(roleElement);
            }

            // now we iterate over roles and add\import the referenced
            // interface components

            for (Role role : application.getRoles()) {

                QName interfaceName = role.getReferencedComponentName();
                String tns = def.getTargetNamespace();
                String cns = interfaceName.getNamespaceURI();

                if (cns.equals(tns)) {

                    // interface is defined within same namespace as this
                    // application add interface declaration to current def
                    // if it doesn't already exist

                    interfaceDefinitionWriter.writeDefinition(def, role.getReferencedComponent());

                } else {

                    // interface is defined within a different namespace. add
                    // fragmentID to wsdlLocation cached on import element

                    ImportElement importElement = (ImportElement)def.getImportElements().get(cns);
                    if (importElement == null) {
                        importElement = (ImportElement)wsdlDocument.createImportElement();
                        importElement.setNamespace(cns);
                        def.addImportElement(importElement);
                        WSDLLocation base = WSDLLocation.valueOf(wsdlDocument.getDocumentURI());
                        WSDLLocation wsdlLocation = new WSDLLocation(base);
                        importElement.setWsdlLocation(wsdlLocation);
                    }

                    WSDLLocation wsdlLocation = importElement.getWsdlLocation();
                    FragmentIdentifier fid = FragmentIdentifier.valueOf(ComponentType.INTERFACE, interfaceName);
                    wsdlLocation.addFragmentIdentifier(fid);

                }

            }

            log.trace("writeDefinition end");

        }

    }

    private static class BindingDefinitionWriter {

        public void writeDefinition(DefinitionsElement def, Binding binding) throws IDLException {

            if (def.getBindingElements().get(binding.getName()) != null) {
                return;
            }

            log.trace("writeDefinition begin");

            // first we verify that this binding is describable via wsdl 1.1

            boolean isWSDL11Binding = false;

            if (binding.getBindingType().equals(SOAPBinding.BINDING_TYPE)) {
                if (((SOAPBinding)binding).getSoapVersion() == SOAPVersion.SOAP11) {
                    if (binding.getUnderlyingProtocol() instanceof SOAP11HTTPProtocol) {
                        isWSDL11Binding = true;
                    }
                }
            }

            if (!isWSDL11Binding) {
                throw new WSDLException("Binding is not describable via WSDL11. Expected type Ref '"
                        + SOAPBinding.BINDING_TYPE
                        + "'. Expected SOAPVersion "
                        + SOAPVersion.SOAP11
                        + ". Expected instanceof '"
                        + SOAP11HTTPProtocol.class.getName()
                        + "' as underlying protocol.");
            }

            // using binding type ref, retreive a factory which understands binding
            // type and can serialize binding component as an element instance. note
            // that the implementation is arbitrary, i.e. impls for the same binding
            // type are extension factory providers for the same binding factory
            // instance

            BindingFactory bindingFactory = null;
            for (BindingFactory tmp : SystemContext.getContext().getSystem().getBindingFactories()) {
                if(tmp.getBindingType().equals(SOAPBinding.BINDING_TYPE)) {
                    bindingFactory = tmp;
                    break;
                }
            }
            if (bindingFactory == null) {
                throw new WSDLException("Error writing definition for Binding "
                        + binding.getName()
                        + ". No BindingFactory mapped to binding type '"
                        + SOAPBinding.BINDING_TYPE
                        + "'.");
            } else {
                bindingFactory.writeBinding((WSDLDocument)def.getOwnerDocument(), binding);
            }

            // now we add\import the referenced interface component

            Interface intrface = binding.getInterfaceReference().getReferencedComponent();
            String tns = def.getTargetNamespace();
            String cns = intrface.getName().getNamespaceURI();

            if (cns.equals(tns)) {

                // interface is defined within same namespace as this
                // binding add interface declaration to current definition

                interfaceDefinitionWriter.writeDefinition(def, intrface);

            } else {

                // interface is defined within a different namespace. add
                // fragmentID to wsdlLocation cached on import element

                ImportElement importElement = (ImportElement)def.getImportElements().get(cns);
                if (importElement == null) {
                    WSDL11Document wsdlDocument = (WSDL11Document)def.getOwnerDocument();
                    importElement = (ImportElement)wsdlDocument.createImportElement();
                    importElement.setNamespace(cns);
                    def.addImportElement(importElement);
                    WSDLLocation base = WSDLLocation.valueOf(wsdlDocument.getDocumentURI());
                    WSDLLocation wsdlLocation = new WSDLLocation(base);
                    importElement.setWsdlLocation(wsdlLocation);
                }

                WSDLLocation wsdlLocation = importElement.getWsdlLocation();
                FragmentIdentifier fid = FragmentIdentifier.valueOf(ComponentType.INTERFACE, intrface.getName());
                wsdlLocation.addFragmentIdentifier(fid);

            }

            log.trace("writeDefinition end");

        }

    }

    private static class EngineDefinitionWriter {

        public void writeDefinition(DefinitionsElement def, Engine engine) throws IDLException {

            if (def.getEngineElements().get(engine.getName()) != null) {
                return;
            }

            log.trace("writeDefinition begin");

            // first we add the engine element to the definition

            WSDL11Document wsdlDocument = (WSDL11Document)def.getOwnerDocument();
            EngineElement engineElement = wsdlDocument.createEngineElement();
            def.addEngineElement(engineElement);
            engineElement.setName(engine.getName().getLocalPart());

            engineElement.setApplication(engine.getApplicationReference().getReferencedComponentName());

            for (Actor actor : engine.getActors()) {
                ActorElement actorElement = wsdlDocument.createActorElement();
                actorElement.setRole(actor.getRoleName());
                actorElement.setService(actor.getReferencedComponentName());
            }

            // now we add\import all of the referenced components

            Application application = engine.getApplicationReference().getReferencedComponent();
            String tns = def.getTargetNamespace();
            String cns = application.getName().getNamespaceURI();

            if (cns.equals(tns)) {

                // application is defined within same namespace as this
                // engine add application declaration to current def

                applicationDefinitionWriter.writeDefinition(def, application);

            } else {

                // application is defined within a different namespace. add
                // fragmentID to wsdlLocation cached on import element

                ImportElement importElement = (ImportElement)def.getImportElements().get(cns);
                if (importElement == null) {
                    importElement = (ImportElement)wsdlDocument.createImportElement();
                    importElement.setNamespace(cns);
                    def.addImportElement(importElement);
                    WSDLLocation base = WSDLLocation.valueOf(wsdlDocument.getDocumentURI());
                    WSDLLocation wsdlLocation = new WSDLLocation(base);
                    importElement.setWsdlLocation(wsdlLocation);
                }

                WSDLLocation wsdlLocation = importElement.getWsdlLocation();
                FragmentIdentifier fid = FragmentIdentifier.valueOf(ComponentType.APPLICATION, application
                        .getName());
                wsdlLocation.addFragmentIdentifier(fid);

            }

            /** ******************* process the services ************************* */

            for (Actor actor : engine.getActors()) {

                Service service = actor.getReferencedComponent();
                cns = service.getName().getNamespaceURI();

                if (cns.equals(tns)) {

                    // service is defined within same namespace as this
                    // engine add service declaration to current def

                    serviceDefinitionWriter.writeDefinition(def, service);

                } else {

                    // service is defined within a different namespace. add
                    // fragmentID to wsdlLocation cached on import element

                    ImportElement importElement = (ImportElement)def.getImportElements().get(cns);
                    if (importElement == null) {
                        importElement = (ImportElement)wsdlDocument.createImportElement();
                        importElement.setNamespace(cns);
                        def.addImportElement(importElement);
                        WSDLLocation base = WSDLLocation.valueOf(wsdlDocument.getDocumentURI());
                        WSDLLocation wsdlLocation = new WSDLLocation(base);
                        importElement.setWsdlLocation(wsdlLocation);
                    }

                    WSDLLocation wsdlLocation = importElement.getWsdlLocation();
                    FragmentIdentifier fid = FragmentIdentifier.valueOf(ComponentType.SERVICE, service.getName());
                    wsdlLocation.addFragmentIdentifier(fid);

                }

            }

            log.trace("writeDefinition end");

        }

    }

    private static class InterfaceDefinitionWriter {

        public void writeDefinition(DefinitionsElement def, Interface intrface) throws WSDLException {

            if (def.getPortTypeElements().get(intrface.getName()) != null) {
                return;
            }

            log.trace("writeDefinition begin");

            if (intrface.getExtendedInterfaces().size() > 0) {
                throw new WSDLException("Definition for Interface "
                        + intrface.getName()
                        + " cannot be exported via WSDL 1.1."
                        + " Interface definition extends from one or more ancestors.");
            }

            if (intrface.getStyleDefaults().size() > 0) {
                throw new WSDLException("Definition for Interface "
                        + intrface.getName()
                        + " cannot be exported via WSDL 1.1."
                        + " Interface definition defines one or more style iri's.");
            }

            // first we add the portType element to the definition and all contained
            // operation elements

            WSDL11Document wsdlDocument = (WSDL11Document)def.getOwnerDocument();

            QName interfaceName = intrface.getName();

            PortTypeElement portType = wsdlDocument.createPortTypeElement();
            def.addPortTypeElement(portType);
            portType.setName(interfaceName.getLocalPart());

            if (intrface.getElementPolicy() != null) {
                PolicyExpression policy;
                try {
                    policy = intrface.getElementPolicy().getMergedExpression();
                } catch (PolicyException pe) {
                    throw new WSDLException("Error writing definition. " + pe);
                }
                PolicyUtil.writePolicy(policy, portType);
            }

            for (InterfaceOperation interfaceOperation : intrface.getDeclaredOperations()) {

                if (interfaceOperation.getMessageExchangePattern() == MEP.ROBUST_IN_ONLY) {
                    throw new WSDLException("Error writing definition. InterfaceOperation "
                            + interfaceOperation.getName()
                            + " indicates 'robust-in-only' MEP, which is not describable via"
                            + " WSDL 1.1");
                }

                if (interfaceOperation.getStyles().size() > 0) {
                    throw new WSDLException("Error writing definition. InterfaceOperation "
                            + interfaceOperation.getName()
                            + " indicates at least one style iri, which is not describable via"
                            + " WSDL 1.1");
                }

                PortTypeOperationElement operationElement = wsdlDocument.createPortTypeOperationElement();
                operationElement.setName(interfaceOperation.getName().getLocalPart());

                if (interfaceOperation.getElementPolicy() != null) {
                    PolicyExpression policy;
                    try {
                        policy = interfaceOperation.getElementPolicy().getMergedExpression();
                    } catch (PolicyException pe) {
                        throw new WSDLException("Error writing definition. " + pe);
                    }
                    PolicyUtil.writePolicy(policy, operationElement);
                }

                // process input message for this operation
                PortTypeOperationInputElement inputElement = wsdlDocument.createPortTypeOperationInputElement();
                operationElement.setInputElement(inputElement);
                InterfaceMessageReference inputMessageRef = interfaceOperation.getInputMessageReference();
                inputElement.setMessage(inputMessageRef.getReferencedComponentName());
                inputElement.setAttributeNS(WSA_METADATA_NS, "wsa:" + ATT_WSA_ACTION, inputMessageRef.getAction());

                if (inputMessageRef.getElementPolicy() != null) {
                    PolicyExpression policy;
                    try {
                        policy = inputMessageRef.getElementPolicy().getMergedExpression();
                    } catch (PolicyException pe) {
                        throw new WSDLException("Error writing definition. " + pe);
                    }
                    PolicyUtil.writePolicy(policy, inputElement);
                }

                // process optional output message for this operation
                InterfaceMessageReference outputMessageRef = interfaceOperation.getOutputMessageReference();
                if (outputMessageRef != null) {
                    PortTypeOperationOutputElement outputElement = wsdlDocument
                            .createPortTypeOperationOutputElement();
                    outputElement.setMessage(outputMessageRef.getReferencedComponentName());
                    operationElement.setOutputElement(outputElement);
                    outputElement.setAttributeNS(WSA_METADATA_NS, "wsa:" + ATT_WSA_ACTION, outputMessageRef
                            .getAction());
                    if (outputMessageRef.getElementPolicy() != null) {
                        PolicyExpression policy;
                        try {
                            policy = outputMessageRef.getElementPolicy().getMergedExpression();
                        } catch (PolicyException pe) {
                            throw new WSDLException("Error writing definition. " + pe);
                        }
                        PolicyUtil.writePolicy(policy, outputElement);
                    }
                }

                // process optional faults for this operation
                for (InterfaceFaultReference interfaceFaultRef : interfaceOperation.getFaultReferences()) {
                    InterfaceFault interfaceFault = interfaceFaultRef.getFault();
                    PortTypeOperationFaultElement faultElement = wsdlDocument
                            .createPortTypeOperationFaultElement();
                    faultElement.setName(interfaceFault.getName().getLocalPart());
                    faultElement.setMessage(interfaceFault.getReferencedComponentName());
                    faultElement.setAttributeNS(WSA_METADATA_NS, "wsa:" + ATT_WSA_ACTION, interfaceFaultRef
                            .getAction());
                    operationElement.addFaultElement(faultElement);
                    if (interfaceFaultRef.getElementPolicy() != null) {
                        PolicyExpression policy;
                        try {
                            policy = interfaceFaultRef.getElementPolicy().getMergedExpression();
                        } catch (PolicyException pe) {
                            throw new WSDLException("Error writing definition. " + pe);
                        }
                        PolicyUtil.writePolicy(policy, faultElement);
                    }
                }

                // add operation to portType
                portType.addOperationElement(operationElement);

            }

            // now we iterate over operations and add\import the referenced
            // message components

            for (InterfaceOperation interfaceOperation : intrface.getDeclaredOperations()) {

                // ********** process request message for this operation ****************

                InterfaceMessage im = interfaceOperation.getInputMessageReference().getReferencedComponent();
                String tns = def.getTargetNamespace();
                String cns = im.getName().getNamespaceURI();

                if (cns.equals(tns)) {

                    // message is defined within same namespace as this interface.
                    // add message declaration to current definition

                    messageDefinitionWriter.writeDefinition(def, im);

                } else {

                    // message is defined within a different namespace. add
                    // fragmentID to wsdlLocation cached on import element

                    ImportElement importElement = (ImportElement)def.getImportElements().get(cns);
                    if (importElement == null) {
                        importElement = (ImportElement)wsdlDocument.createImportElement();
                        importElement.setNamespace(cns);
                        def.addImportElement(importElement);
                        WSDLLocation base = WSDLLocation.valueOf(wsdlDocument.getDocumentURI());
                        WSDLLocation wsdlLocation = new WSDLLocation(base);
                        importElement.setWsdlLocation(wsdlLocation);
                    }

                    WSDLLocation wsdlLocation = importElement.getWsdlLocation();
                    FragmentIdentifier fid = FragmentIdentifier.valueOf(ComponentType.MESSAGE, im.getName());
                    wsdlLocation.addFragmentIdentifier(fid);

                }

                // *********** process response message for this operation *************

                if (interfaceOperation.getOutputMessageReference() != null) {

                    InterfaceMessage om = interfaceOperation.getOutputMessageReference().getReferencedComponent();
                    cns = om.getName().getNamespaceURI();

                    if (cns.equals(tns)) {

                        // message is defined within same namespace as this interface.
                        // add message declaration to current definition

                        messageDefinitionWriter.writeDefinition(def, om);

                    } else {

                        // message is defined within a different namespace. add
                        // fragmentID to wsdlLocation cached on import element

                        ImportElement importElement = (ImportElement)def.getImportElements().get(cns);
                        if (importElement == null) {
                            importElement = (ImportElement)wsdlDocument.createImportElement();
                            importElement.setNamespace(cns);
                            def.addImportElement(importElement);
                            WSDLLocation base = WSDLLocation.valueOf(wsdlDocument.getDocumentURI());
                            WSDLLocation wsdlLocation = new WSDLLocation(base);
                            importElement.setWsdlLocation(wsdlLocation);
                        }

                        WSDLLocation wsdlLocation = importElement.getWsdlLocation();
                        FragmentIdentifier fid = FragmentIdentifier.valueOf(ComponentType.MESSAGE, om.getName());
                        wsdlLocation.addFragmentIdentifier(fid);

                    }
                }

                // *********** process fault messages for this operation *************

                for (InterfaceFaultReference interfaceFaultRef : interfaceOperation.getFaultReferences()) {

                    InterfaceFault interfaceFault = interfaceFaultRef.getFault();
                    InterfaceMessage fm = interfaceFault.getReferencedComponent();
                    cns = fm.getName().getNamespaceURI();

                    if (cns.equals(tns)) {

                        // message is defined within same namespace as this interface.
                        // add message declaration to current definition

                        messageDefinitionWriter.writeDefinition(def, fm);

                    } else {

                        // message is defined within a different namespace. add
                        // fragmentID to wsdlLocation cached on import element

                        ImportElement importElement = (ImportElement)def.getImportElements().get(cns);
                        if (importElement == null) {
                            importElement = (ImportElement)wsdlDocument.createImportElement();
                            importElement.setNamespace(cns);
                            def.addImportElement(importElement);
                            WSDLLocation base = WSDLLocation.valueOf(wsdlDocument.getDocumentURI());
                            WSDLLocation wsdlLocation = new WSDLLocation(base);
                            importElement.setWsdlLocation(wsdlLocation);
                        }

                        WSDLLocation wsdlLocation = importElement.getWsdlLocation();
                        FragmentIdentifier fid = FragmentIdentifier.valueOf(ComponentType.MESSAGE, fm.getName());
                        wsdlLocation.addFragmentIdentifier(fid);

                    }
                }
            }

            log.trace("writeDefinition end");

        }

    }

    private static class MessageDefinitionWriter {

        public void writeDefinition(DefinitionsElement def, InterfaceMessage message) throws WSDLException {

            if (def.getMessageElements().get(message.getName()) != null) {
                return;
            }

            log.trace("writeDefinition begin");

            WSDL11Document wsdlDocument = (WSDL11Document)def.getOwnerDocument();
            QName messageName = message.getName();
            MessageElement messageElement = wsdlDocument.createMessageElement();
            def.addMessageElement(messageElement);
            messageElement.setName(messageName.getLocalPart());

            if (message.getElementPolicy() != null) {
                PolicyExpression elementPolicy;
                try {
                    elementPolicy = message.getElementPolicy().getMergedExpression();
                } catch (PolicyException pe) {
                    throw new WSDLException("Error writing definition. " + pe);
                }
                PolicyUtil.writePolicy(elementPolicy, messageElement);
            }

            for (MessagePart abstractPart : message.getParts()) {
                MessagePartElement partElement = wsdlDocument.createMessagePartElement();
                partElement.setName(abstractPart.getName());
                QName schemaComponentName = abstractPart.getSchemaComponentQName();
                if (abstractPart.getSchemaComponentType().equals(SchemaComponent.Type.ELEMENT_DECLARATION)) {
                    partElement.setElement(schemaComponentName);
                } else {
                    partElement.setType(schemaComponentName);
                }
                messageElement.addPartElement(partElement);
                if (log.isDebugEnabled()) {
                    log.debug("added part " + abstractPart.getName() + ".");
                }
            }

            // now iterate over message parts and add referenced schemas to
            // types section

            for (MessagePart messagePart : message.getParts()) {
                importSchema(def, messagePart.getSchemaComponentNamespace());
            }

            log.trace("writeDefinition end");

        }

        private static void importSchema(DefinitionsElement def, String targetNamespace) throws WSDLException {
            log.trace("importSchema begin");

            WSDL11Document wsdlDocument = (WSDL11Document)def.getOwnerDocument();

            TypesElement typesElement = def.getTypesElement();
            if (typesElement == null) {
                typesElement = wsdlDocument.createTypesElement();
                def.setTypesElement(typesElement);
            }

            // by convention, types element will contain a single schema child element
            // with no targetnamespace. we'll add all referenced schemas as a child
            // import element of 'shell' schema element. location will resolve to
            // wsdl server feature on this machine

            boolean exists = false;

            Element schemaElement = DOMUtil.getChildElement(typesElement, new QName(XMLSchema.TYPE, "schema"));
            if (schemaElement == null) {
                schemaElement = wsdlDocument.createElementNS(XMLSchema.TYPE, "xs:schema");
                typesElement.appendChild(schemaElement);
            }

            List<Element> importElements = DOMUtil.getChildElements(schemaElement, new QName(XMLSchema.TYPE,
                    "import"));
            for (Element importElement : importElements) {
                if (importElement.getAttribute("namespace").equals(targetNamespace)) {
                    exists = true;
                    break;
                }
            }

            if (!exists) {
                WSDLLocation baseLocation = WSDLLocation.valueOf(wsdlDocument.getBaseURI());
                WSDLLocation wsdlLocation = new WSDLLocation(baseLocation);
                QName schemaName = new QName(targetNamespace, "");
                FragmentIdentifier fid = FragmentIdentifier.valueOf(ComponentType.SCHEMA, schemaName);
                wsdlLocation.addFragmentIdentifier(fid);
                Element importElement = wsdlDocument.createElementNS(XMLSchema.TYPE, "xs:import");
                schemaElement.appendChild(importElement);
                importElement.setAttribute("namespace", targetNamespace);
                importElement.setAttribute("location", wsdlLocation.toString());
            }

            log.trace("importSchema end");
        }

    }

    private static class ServiceDefinitionWriter {

        public void writeDefinition(DefinitionsElement def, Service service) throws IDLException {

            if (def.getServiceElements().get(service.getName()) != null) {
                return;
            }

            log.trace("writeDefinition begin");

            // first we add the service element to the definition

            WSDL11Document wsdlDocument = (WSDL11Document)def.getOwnerDocument();

            QName serviceName = service.getName();
            ServiceElement serviceElement = wsdlDocument.createServiceElement();
            serviceElement.setName(serviceName.getLocalPart());
            def.addServiceElement(serviceElement);

            if (service.getElementPolicy() != null) {
                PolicyExpression elementPolicy = null;
                try {
                    elementPolicy = service.getElementPolicy().getMergedExpression();
                } catch (PolicyException pe) {
                    throw new WSDLException("Error writing definition. " + pe);
                }
                PolicyUtil.writePolicy(elementPolicy, serviceElement);
            }

            for (Endpoint endpoint : service.getEndpoints()) {
                ServicePortElement portElement = wsdlDocument.createServicePortElement();
                portElement.setName(endpoint.getName());
                serviceElement.addPortElement(portElement);
                Binding binding = endpoint.getBindingReference().getReferencedComponent();
                portElement.setBinding(binding.getName());
                SOAPAddressElement soapAddressElement = null;
                if (soapAddressElement instanceof SOAP11AddressElement) {
                    soapAddressElement = wsdlDocument.createSOAP11AddressElement();
                } else {
                    soapAddressElement = wsdlDocument.createSOAP12AddressElement();
                }
                soapAddressElement.setLocation(endpoint.getAddress());
                if (endpoint.getElementPolicy() != null) {
                    PolicyExpression elementPolicy = null;
                    try {
                        elementPolicy = endpoint.getElementPolicy().getMergedExpression();
                    } catch (PolicyException pe) {
                        throw new WSDLException("Error writing definition. " + pe);
                    }
                    PolicyUtil.writePolicy(elementPolicy, portElement);
                }
            }

            // now we add\import all of the referenced components

            for (Endpoint endpoint : service.getEndpoints()) {

                Binding binding = endpoint.getBindingReference().getReferencedComponent();
                String tns = def.getTargetNamespace();
                String cns = binding.getName().getNamespaceURI();

                if (cns.equals(tns)) {

                    // binding is defined within same namespace as this
                    // service add binding declaration to current def

                    bindingDefinitionWriter.writeDefinition(def, binding);

                } else {

                    // binding is defined within a different namespace. add
                    // fragmentID to wsdlLocation cached on import element

                    ImportElement importElement = (ImportElement)def.getImportElements().get(cns);
                    if (importElement == null) {
                        importElement = (ImportElement)wsdlDocument.createImportElement();
                        importElement.setNamespace(cns);
                        def.addImportElement(importElement);
                        WSDLLocation base = WSDLLocation.valueOf(wsdlDocument.getDocumentURI());
                        WSDLLocation wsdlLocation = new WSDLLocation(base);
                        importElement.setWsdlLocation(wsdlLocation);
                    }

                    WSDLLocation wsdlLocation = importElement.getWsdlLocation();
                    FragmentIdentifier fid = FragmentIdentifier.valueOf(ComponentType.BINDING, binding.getName());
                    wsdlLocation.addFragmentIdentifier(fid);

                }

            }

            log.trace("writeDefinition end");

        }

    }

    /** ****************************** Utility Classes *********************************** * */

    private static class BasicProfile {

        private static BindingFaultValidator bindingFaultValidator = new BindingFaultValidator();
        private static BindingInputValidator bindingInputValidator = new BindingInputValidator();
        private static BindingOperationValidator bindingOperationValidator = new BindingOperationValidator();
        private static BindingOutputValidator bindingOutputValidator = new BindingOutputValidator();
        private static BindingValidator bindingValidator = new BindingValidator();
        private static MessageValidator messageValidator = new MessageValidator();
        private static PortTypeFaultValidator portTypeFaultValidator = new PortTypeFaultValidator();
        private static PortTypeInputValidator portTypeInputValidator = new PortTypeInputValidator();
        private static PortTypeOperationValidator portTypeOperationValidator = new PortTypeOperationValidator();
        private static PortTypeOutputValidator portTypeOutputValidator = new PortTypeOutputValidator();
        private static PortTypeValidator portTypeValidator = new PortTypeValidator();
        private static ServiceValidator serviceValidator = new ServiceValidator();
        private static SOAPBindingValidator soapBindingValidator = new SOAPBindingValidator();
        private static SOAPOperationValidator soapOperationValidator = new SOAPOperationValidator();
        private static TypesValidator typesValidator = new TypesValidator();
        private static TypeValidator typeValidator = new TypeValidator();

        public static ValidationInfo validatePortType(DefinitionsElement def, QName componentName) {
            log.trace("validatePortType begin");

            ValidationInfo validationInfo = new ValidationInfo();
            ValidationMessage validationMessage = new ValidationMessage();
            validationInfo.setValidationMessage(validationMessage);

            TypesElement typesElement = def.getTypesElement();
            if (typesElement != null) {
                typesValidator.validate(typesElement, validationInfo);
            }

            PortTypeElement portType = def.getPortTypeElements().get(componentName.getLocalPart());
            if (portType == null) {
                validationMessage.append("No portType found within def named " + componentName.getLocalPart());
            } else {
                if (portType.isValid() == null) {
                    portTypeValidator.validate(def, portType, validationInfo);
                    if (validationInfo.getValidationMessage().length() > 0) {
                        portType.setIsValid(Boolean.FALSE);
                    } else {
                        portType.setIsValid(Boolean.TRUE);
                    }
                }
            }

            log.trace("validatePortType end");
            return validationInfo;
        }

        public static ValidationInfo validateBinding(DefinitionsElement def, QName componentName) {
            log.trace("validateBinding begin");

            ValidationInfo validationInfo = new ValidationInfo();
            ValidationMessage validationMessage = new ValidationMessage();
            validationInfo.setValidationMessage(validationMessage);

            TypesElement typesElement = def.getTypesElement();
            if (typesElement != null) {
                typesValidator.validate(typesElement, validationInfo);
            }

            BindingElement binding = def.getBindingElements().get(componentName.getLocalPart());
            if (binding == null) {
                validationMessage.append("No binding found within def named " + componentName.getLocalPart());
            } else {
                if (binding.isValid() == null) {
                    bindingValidator.validate(def, binding, validationInfo);
                    if (validationInfo.getValidationMessage().length() > 0) {
                        binding.setIsValid(Boolean.FALSE);
                    } else {
                        binding.setIsValid(Boolean.TRUE);
                    }
                }
            }

            log.trace("validateBinding end");
            return validationInfo;
        }

        public static ValidationInfo validateService(DefinitionsElement def, QName componentName) {
            log.trace("validateService begin");

            ValidationInfo validationInfo = new ValidationInfo();
            ValidationMessage validationMessage = new ValidationMessage();
            validationInfo.setValidationMessage(validationMessage);

            TypesElement typesElement = def.getTypesElement();
            if (typesElement != null) {
                typesValidator.validate(typesElement, validationInfo);
            }

            ServiceElement serviceElement = def.getServiceElements().get(componentName.getLocalPart());
            if (serviceElement == null) {
                validationMessage.append("No service found within def named " + componentName.getLocalPart());
            } else {
                if (serviceElement.isValid() == null) {
                    serviceValidator.validate(def, serviceElement, validationInfo);
                    if (validationInfo.getValidationMessage().length() > 0) {
                        serviceElement.setIsValid(Boolean.FALSE);
                    } else {
                        serviceElement.setIsValid(Boolean.TRUE);
                    }
                }
            }

            log.trace("validateService end");
            return validationInfo;
        }

        public static ValidationInfo validateMessage(DefinitionsElement def, QName componentName) {
            log.trace("validateMessage begin");

            ValidationInfo validationInfo = new ValidationInfo();
            ValidationMessage validationMessage = new ValidationMessage();
            validationInfo.setValidationMessage(validationMessage);

            TypesElement typesElement = def.getTypesElement();
            if (typesElement != null) {
                typesValidator.validate(typesElement, validationInfo);
            }

            MessageElement message = def.getMessageElements().get(componentName.getLocalPart());
            if (message == null) {
                validationMessage.append("No message found within def named " + componentName.getLocalPart());
            } else {
                if (message.isValid() == null) {
                    messageValidator.validate(message, validationInfo);
                    if (validationInfo.getValidationMessage().length() > 0) {
                        message.setIsValid(Boolean.FALSE);
                    } else {
                        message.setIsValid(Boolean.TRUE);
                    }
                }
            }

            log.trace("validateMessage end");
            return validationInfo;
        }

        private static class BindingFaultValidator {

            private static final String EMPTY_STRING = "";

            public void validate(DefinitionsElement def, BindingElement binding, BindingOperationElement bindingOperation, BindingOperationFaultElement bindingOperationFault, ValidationInfo validationInfo) {

                log.trace("validate begin");

                QName bindingName = new QName(def.getTargetNamespace(), binding.getName());
                String operationName = bindingOperation.getName();

                ValidationMessage validationMessage = validationInfo.getValidationMessage();

                if (bindingOperationFault.getName() == null
                        || bindingOperationFault.getName().equals(EMPTY_STRING)) {
                    validationMessage.append("Fault name invalid within binding operation '"
                            + operationName
                            + "'."
                            + " A wsdl:binding in a DESCRIPTION MUST have the name attribute specified"
                            + " on all contained soapbind:fault elements."
                            + " (R2721 ws-i basic profile 1.1)");
                }

                SOAPFaultElement soapFaultElement = bindingOperationFault.getSOAPFaultElement();

                if (soapFaultElement == null) {
                    validationMessage.append("No soap:fault declared within soapbind:fault in binding "
                            + bindingName
                            + ".");
                } else {

                    if (soapFaultElement.getNamespace() != null
                            && validationInfo.getBindingStyle() == BindingStyle.RPC_STYLE) {
                        validationMessage.append("soap:fault element in binding:operation "
                                + operationName
                                + " defined on Binding "
                                + bindingName
                                + " is invalid."
                                + " An rpc-literal binding in a DESCRIPTION MUST NOT have the namespace"
                                + " attribute specified on contained soapbind:header, soapbind:headerfault"
                                + " and soapbind:fault elements."
                                + " (R2726 ws-i basic profile 1.1)");
                    } else if (soapFaultElement.getNamespace() != null) {
                        validationMessage.append("soap:fault element in binding:operation "
                                + operationName
                                + " defined on Binding "
                                + bindingName
                                + " is invalid."
                                + " A document-literal binding in a DESCRIPTION MUST NOT have the namespace"
                                + " attribute specified on contained soapbind:body, soapbind:header,"
                                + " soapbind:headerfault and soapbind:fault elements."
                                + " (R2716 ws-i basic profile 1.1)");
                    }

                    if (soapFaultElement.getUse().equals(ENCODED_USE)) {
                        validationMessage.append("soap:fault declared within soapbind:fault in binding "
                                + bindingName
                                + " is invalid."
                                + " A wsdl:binding in a DESCRIPTION MUST use the value of"
                                + " 'literal' for the use attribute in all soapbind:body,"
                                + " soapbind:fault, soapbind:header and soapbind:headerfault"
                                + " elements."
                                + " (R2706 ws-i basic profile 1.1)");
                    } else {
                        // spec allows a default value of 'literal' (see R2707)
                    }
                }

                // break out at this point if any validation errors reported, as the
                // logic below relies upon correct declaration of soap:fault
                if (validationMessage.toString() != null) {
                    log.trace("validate end");
                    return;
                }

                if (!(bindingOperationFault.getName().equals(soapFaultElement.getName()))) {
                    validationMessage.append("name attribute on soapbind:fault '"
                            + bindingOperationFault.getName()
                            + "' defined on Binding "
                            + bindingName
                            + " is invalid."
                            + " In a DESCRIPTION, the value of the name attribute on a soapbind:fault"
                            + " element MUST match the value of the name attribute on its parent"
                            + " wsdl:fault element"
                            + " (R2754 ws-i basic profile 1.1)");
                }

                // retrieve portType element, which we know exists because it was resolved by
                // binding validator

                QName portTypeName = binding.getType();

                PortTypeElement portType = null;
                if (portTypeName.getNamespaceURI().equals(def.getTargetNamespace())) {
                    portType = def.getPortTypeElements().get(portTypeName.getLocalPart());
                } else {
                    ImportElement importElement = def.getImportElements().get(portTypeName.getNamespaceURI());
                    if (importElement != null) {
                        DefinitionsElement temp = null;
                        try {
                            temp = importElement.getDocument().getDefinitionsElement();
                        } catch (WSDLException fatchance) {
                        }
                        portType = temp.getPortTypeElements().get(portTypeName.getLocalPart());
                    }
                }

                // using faultName retrieve message from wsdl:operation in portType. we know op
                // exists because it was resolved by binding validator and message existence
                // was validated by port type validator

                MessageElement faultMessageElement = null;

                PortTypeOperationElement portTypeOperation = portType.getOperationElements().get(operationName);
                PortTypeOperationFaultElement portTypeOperationFaultElement = portTypeOperation.getFaultElements()
                        .get(bindingOperationFault.getName());
                if (portTypeOperationFaultElement == null) {
                    validationMessage.append("soap:fault declared within soapbind:fault '"
                            + bindingOperationFault.getName()
                            + "' defined on Binding "
                            + bindingName
                            + " has no corresponding fault declared in wsdl:operation '"
                            + portTypeOperation.getName()
                            + "'.");
                    log.trace("validate end");
                    return;
                }

                QName faultMessageName = portTypeOperationFaultElement.getMessage();
                if (faultMessageName.getNamespaceURI().equals(def.getTargetNamespace())) {
                    faultMessageElement = def.getMessageElements().get(faultMessageName.getLocalPart());
                } else {
                    ImportElement importElement = def.getImportElements().get(faultMessageName.getNamespaceURI());
                    if (importElement != null) {
                        DefinitionsElement temp = null;
                        try {
                            temp = importElement.getDocument().getDefinitionsElement();
                        } catch (WSDLException fatchance) {
                        }
                        faultMessageElement = temp.getMessageElements().get(faultMessageName.getLocalPart());
                    }
                }

                // a fault message, ie. a message referenced by a wsdl:fault, may only
                // contain a single part per wsdl 1.1 spec (section 3.6). also, wsdl
                // semantics preclude declaration of soap headers on a fault

                MessagePartElement faultPartElement = null;

                if (faultMessageElement.getPartElements().values().size() != 1) {
                    validationMessage.append("Binding "
                            + bindingName
                            + " is invalid. Fault message (ie. a message referenced by a wsdl:fault) "
                            + faultMessageName
                            + " must contain exactly one part "
                            + " in accordance with wsdl 1.1 spec. sec. 3.6");
                } else {
                    faultPartElement = faultMessageElement.getPartElements().values().iterator().next();
                }

                if (faultPartElement.getElement() == null) {
                    validationMessage.append("wsdl:part '"
                            + faultPartElement.getName()
                            + "' referenced by fault "
                            + portTypeOperationFaultElement.getName()
                            + " declared on operation '"
                            + operationName
                            + " defined on Binding "
                            + bindingName
                            + " is invalid."
                            + " A wsdl:binding in a DESCRIPTION MUST refer, in each of its soapbind:header,"
                            + " soapbind:headerfault and soapbind:fault elements, only to wsdl:part element(s)"
                            + " that have been defined using the element attribute."
                            + " (R2205 ws-i basic profile 1.1)");
                }

                log.trace("validate end");

            }

        }

        private static class BindingInputValidator {

            public QName validate(DefinitionsElement def, BindingOperationElement bindingOperation, PortTypeOperationElement portTypeOperation, BindingOperationInputElement bindingInput, ValidationInfo validationInfo) {

                log.trace("validate begin");

                ValidationMessage validationMessage = validationInfo.getValidationMessage();

                SOAPBodyElement soapBody = bindingInput.getSOAPBodyElement();
                Map<String, SOAPHeaderElement> soapHeaders = new HashMap<String, SOAPHeaderElement>();

                QName soapBodyType = null;

                if (soapBody == null) {
                    validationMessage.append("wsdl:input element in binding operation "
                            + bindingOperation.getName()
                            + " has no soap:body");
                    return null;
                } else {

                    if (portTypeOperation.getInputElement() == null) {

                        validationMessage.append("No input defined for operation " + portTypeOperation.getName());
                        return null;

                    } else {

                        QName messageName = portTypeOperation.getInputElement().getMessage();

                        // wsdl 1.1 (Section 3.7) in regards to headers, indicates "referenced
                        // message need not be the same as the message that defines the SOAP
                        // Body." this breaks bpel 1.1 semantics which associates a bpel
                        // variable with a single wsdl:message type. we will therefore
                        // dissallow
                        // this. a header part must reference same wsdl:message type used by
                        // soap body (which will allows us to index headers by part name

                        for (SOAPHeaderElement soapHeader : bindingInput.getSOAPHeaderElements()) {
                            if (soapHeader.getMessage() == null || !soapHeader.getMessage().equals(messageName)) {
                                validationMessage.append("soap:header in binding:input for binding operation '"
                                        + bindingOperation.getName()
                                        + "' references a message other than 'input' message type "
                                        + messageName
                                        + ". Configuration is not supported.");
                            } else {
                                if (soapHeaders.containsKey(soapHeader.getPart())) {
                                    validationMessage
                                            .append("soap:header in binding:input for binding operation '"
                                                    + bindingOperation.getName()
                                                    + "' binds abstract part '"
                                                    + soapHeader.getPart()
                                                    + "' which is already bound by a header.");
                                } else {
                                    soapHeaders.put(soapHeader.getPart(), soapHeader);
                                }
                            }
                        }

                    }

                    // now we validate that binding input is ws-i compliant. see below
                    // for comments

                    soapBodyType = validateBindingInput(def, bindingOperation, portTypeOperation, soapBody,
                            soapHeaders, validationInfo);
                }

                log.trace("validate end");
                return soapBodyType;
            }

            private QName validateBindingInput(DefinitionsElement def, BindingOperationElement bindingOperation, PortTypeOperationElement portTypeOperation, SOAPBodyElement soapBody, Map<String, SOAPHeaderElement> soapHeaders, ValidationInfo validationInfo) {

                log.trace("validateBindingInput begin");

                ValidationMessage validationMessage = validationInfo.getValidationMessage();

                // validate the soapBody. for a doc-lit type binding if the part attribute on
                // soap:body is absent, then corresponding message may define zero or one part
                // each of which must use the 'element' attribute. if binding is an rpc-lit
                // binding, then zero or more parts may be specified, all of which must
                // reference
                // parts within message using the 'type' attribute. in either case, an empty
                // soap
                // body is valid, ie. no parts defined in either soap:body or wsdl:message.

                if (soapBody.getUse() == null || soapBody.getUse().equals(ENCODED_USE)) {
                    validationMessage.append("soap:body element in wsdl:input element in binding operation "
                            + bindingOperation.getName()
                            + " is invalid."
                            + " A wsdl:binding in a DESCRIPTION MUST use the value of"
                            + " 'literal' for the use attribute in all soapbind:body,"
                            + " soapbind:fault, soapbind:header and soapbind:headerfault"
                            + " elements."
                            + " (R2706 ws-i basic profile 1.1)");
                } else {
                    // spec allows a default value of 'literal' (see R2707)
                }

                // retrieve message element referenced by port type input element. we know
                // it exists because it was resolved by port type input validator

                MessageElement message = null;
                QName messageName = portTypeOperation.getInputElement().getMessage();
                if (messageName.getNamespaceURI().equals(def.getTargetNamespace())) {
                    message = def.getMessageElements().get(messageName.getLocalPart());
                } else {
                    ImportElement importElement = def.getImportElements().get(messageName.getNamespaceURI());
                    if (importElement != null) {
                        DefinitionsElement temp = null;
                        try {
                            temp = importElement.getDocument().getDefinitionsElement();
                        } catch (WSDLException fatchance) {
                        }
                        message = temp.getMessageElements().get(messageName.getLocalPart());
                    }
                }

                Map<String, MessagePartElement> messagePartsMap = message.getPartElements();

                // validate concrete parts within doc-lit binding against the parts
                // within abstract message definition

                QName soapBodyType = null;

                if (validationInfo.getBindingStyle() == BindingStyle.DOCUMENT_STYLE) {

                    List<String> bodyParts = soapBody.getParts();

                    if (soapBody.getNamespace() != null) {
                        validationMessage.append("soap:body element in binding:operation "
                                + bindingOperation.getName()
                                + " is invalid."
                                + " A document-literal binding in a DESCRIPTION MUST NOT have"
                                + " the namespace attribute specified on contained soapbind:body,"
                                + " soapbind:header, soapbind:headerfault and soapbind:fault elements."
                                + " (R2716 ws-i basic profile 1.1)");
                    }

                    QName partType = null;
                    MessagePartElement messagePart = null;
                    if (bodyParts.size() == 0) {
                        if (messagePartsMap.size() > 1) {
                            validationMessage.append("soap:body element in binding:operation "
                                    + bindingOperation.getName()
                                    + " is invalid."
                                    + " If a document-literal binding in a DESCRIPTION does not"
                                    + " specify the parts attribute on a soapbind:body element, the"
                                    + " corresponding abstract wsdl:message MUST define zero or one"
                                    + " wsdl:parts."
                                    + " (R2210 ws-i basic profile 1.1)");
                        }
                        if (messagePartsMap.values().size() == 1) {
                            messagePart = messagePartsMap.values().iterator().next();
                            partType = messagePart.getElement();
                            if (partType == null) {
                                validationMessage.append("wsdl:part '"
                                        + messagePart.getName()
                                        + "' referenced by operation '"
                                        + bindingOperation.getName()
                                        + "' in doc-lit binding is invalid."
                                        + " A document-literal binding in a DESCRIPTION MUST refer, in"
                                        + " each of its soapbind:body element(s), only to wsdl:part"
                                        + " element(s) that have been defined using the element attribute."
                                        + " (R2204 ws-i basic profile 1.1)");
                            }
                        }
                    } else if (bodyParts.size() > 1) {
                        validationMessage.append("soap:body element in wsdl:input element in binding operation '"
                                + bindingOperation.getName()
                                + "' is invalid."
                                + " A document-literal binding in a DESCRIPTION MUST, in each of its"
                                + " soapbind:body element(s), have at most one part listed in the parts"
                                + " attribute, if the parts attribute is specified."
                                + " (R2201 ws-i basic profile 1.1)");

                    } else {
                        String partName = bodyParts.iterator().next();
                        messagePart = messagePartsMap.get(partName);
                        if (messagePart == null) {
                            validationMessage.append("wsdl:input message '"
                                    + messageName
                                    + "' contains no part with name matching name '"
                                    + partName
                                    + "' in corresponding soap:body");
                        } else {
                            partType = messagePart.getElement();
                            if (partType == null) {
                                validationMessage.append("wsdl:part '"
                                        + messagePart.getName()
                                        + "' within input message "
                                        + messageName
                                        + ", referenced by operation '"
                                        + bindingOperation.getName()
                                        + "' in doc-lit binding is invalid."
                                        + " A document-literal binding in a DESCRIPTION MUST refer, in"
                                        + " each of its soapbind:body element(s), only to wsdl:part"
                                        + " element(s) that have been defined using the element attribute."
                                        + " (R2204 ws-i basic profile 1.1)");
                            }
                        }
                    }

                    if (partType != null) {

                        // verify that an element decl exists within schema types for
                        // indicated element type

                        TypesElement typesElement = def.getTypesElement();

                        if (typesElement == null
                                || !typeValidator.validateType(typesElement,
                                        SchemaComponent.Type.ELEMENT_DECLARATION, partType)) {
                            validationMessage.append("wsdl:part '"
                                    + messagePart.getName()
                                    + "' within input message "
                                    + messageName
                                    + ", referenced by operation '"
                                    + bindingOperation.getName()
                                    + "' in doc-lit binding is invalid. No schema found for element "
                                    + partType
                                    + ". A QName reference to a Schema component in a DESCRIPTION MUST"
                                    + " use the namespace defined in the targetNamespace attribute on"
                                    + " the xsd:schema element, or a namespace defined on "
                                    + " namespace attribute on an xsd:import element within the"
                                    + " xsd:schema element."
                                    + " (R2102 ws-i basic profile 1.1)");
                        }

                        soapBodyType = partType;

                    }
                }

                // validate concrete parts within rpc-lit binding against the parts
                // within abstract message definition. the only thing we're looking for
                // here is that if parts are defined, they exist in message and are
                // typed with a 'type' attribute value, ie. not 'element'

                if (validationInfo.getBindingStyle() == BindingStyle.RPC_STYLE) {

                    List<String> bodyParts = soapBody.getParts();

                    if (soapBody.getNamespace() == null) {
                        validationMessage.append("soap:body element in binding:operation "
                                + bindingOperation.getName()
                                + " is invalid."
                                + " An rpc-literal binding in a DESCRIPTION MUST have the namespace"
                                + " attribute specified, the value of which MUST be an absolute URI,"
                                + " on contained soapbind:body elements."
                                + " (R2717 ws-i basic profile 1.1)");
                    }

                    if (bodyParts.size() > 0) {
                        if (messagePartsMap == null || messagePartsMap.size() == 0) {
                            validationMessage.append("wsdl:input message '"
                                    + messageName
                                    + "' contains no parts, but soap:body within input of"
                                    + " soap:operation "
                                    + bindingOperation.getName()
                                    + " specifies at least one part");
                        } else {
                            QName partType = null;
                            MessagePartElement messagePart = null;
                            for (String partName : bodyParts) {
                                messagePart = messagePartsMap.get(partName);
                                if (messagePart == null) {
                                    validationMessage.append("wsdl:input message '"
                                            + messageName
                                            + "' contains no part with name matching name '"
                                            + partName
                                            + "' in corresponding soap:body");
                                } else {
                                    partType = messagePart.getType();
                                    if (partType == null) {
                                        validationMessage.append("wsdl:part '"
                                                + messagePart.getName()
                                                + "' referenced by operation '"
                                                + bindingOperation.getName()
                                                + "' in rpc-lit binding is invalid."
                                                + " A rpc-literal binding in a DESCRIPTION MUST refer, in"
                                                + " each of its soapbind:body element(s), only to wsdl:part"
                                                + " element(s) that have been defined using the type attribute."
                                                + " (R2203 ws-i basic profile 1.1)");
                                    } else {

                                        // verify that an xml type decl exists within schema
                                        // types for indicated xml type, unless type is a
                                        // built-in type in which case system is required to
                                        // load schema by default

                                        if (!partType.getNamespaceURI().equals(XMLSchema.TYPE)) {

                                            TypesElement typesElement = def.getTypesElement();

                                            if (typesElement == null
                                                    || !typeValidator.validateType(typesElement,
                                                            SchemaComponent.Type.TYPE_DEFINITION, partType)) {
                                                validationMessage
                                                        .append("wsdl:part '"
                                                                + messagePart.getName()
                                                                + "' within input message "
                                                                + messageName
                                                                + ", referenced by operation '"
                                                                + bindingOperation.getName()
                                                                + "' in rpc-lit binding is invalid. No schema found for type "
                                                                + partType
                                                                + ". A QName reference to a Schema component in a DESCRIPTION MUST"
                                                                + " use the namespace defined in the targetNamespace attribute on"
                                                                + " the xsd:schema element, or to a namespace defined in the"
                                                                + " namespace attribute on an xsd:import element within the"
                                                                + " xsd:schema element."
                                                                + " (R2102 ws-i basic profile 1.1)");
                                            }

                                        }
                                    }

                                }
                            }
                        }

                    }
                }

                // validate the headers defined for binding input, specifically whether
                // they are mapped to a part in a message and that the part is typed
                // using the element attribute. this is required regardless of binding
                // type, ie rpc and document type

                for (SOAPHeaderElement soapHeader : soapHeaders.values()) {

                    if (soapHeader.getNamespace() != null
                            && validationInfo.getBindingStyle() == BindingStyle.RPC_STYLE) {
                        validationMessage.append("soap:header element in binding:operation "
                                + bindingOperation.getName()
                                + " is invalid."
                                + " An rpc-literal binding in a DESCRIPTION MUST NOT have the namespace"
                                + " attribute specified on contained soapbind:header, soapbind:headerfault"
                                + " and soapbind:fault elements."
                                + " (R2726 ws-i basic profile 1.1)");
                    } else if (soapHeader.getNamespace() != null) {
                        validationMessage.append("soap:header element in binding:operation "
                                + bindingOperation.getName()
                                + " is invalid."
                                + " A document-literal binding in a DESCRIPTION MUST NOT have the namespace"
                                + " attribute specified on contained soapbind:body, soapbind:header,"
                                + " soapbind:headerfault and soapbind:fault elements."
                                + " (R2716 ws-i basic profile 1.1)");
                    }

                    if (soapHeader.getUse() == null || soapHeader.getUse().equals(ENCODED_USE)) {
                        validationMessage.append("soap:header defined in input for binding operation '"
                                + bindingOperation.getName()
                                + "' is invalid."
                                + " A wsdl:binding in a DESCRIPTION MUST use the value of"
                                + " 'literal' for the use attribute in all soapbind:body,"
                                + " soapbind:fault, soapbind:header and soapbind:headerfault"
                                + " elements."
                                + " (R2706 ws-i basic profile 1.1)");
                    }

                    QName headerMessageName = soapHeader.getMessage();

                    MessagePartElement messagePart = messagePartsMap.get(soapHeader.getPart());

                    if (messagePart == null) {
                        validationMessage.append("No part found in message '"
                                + headerMessageName
                                + "' matching part name '"
                                + soapHeader.getPart()
                                + "' referenced by soap:header in input for binding operation '"
                                + bindingOperation.getName()
                                + "'.");
                    }
                    if (messagePart.getElement() == null) {
                        validationMessage.append("Part '"
                                + messagePart.getName()
                                + "' defined in message '"
                                + headerMessageName
                                + "' mapped to soap:header in input for binding operation '"
                                + bindingOperation.getName()
                                + "' is invalid."
                                + " A wsdl:binding in a DESCRIPTION MUST refer, in each of its"
                                + " soapbind:header, soapbind:headerfault and soapbind:fault"
                                + " elements, only to wsdl:part element(s) that have been defined"
                                + " using the element attribute."
                                + " (R2205 ws-i basic profile 1.1)");
                    } else {

                        // verify that an element decl exists within schema types for
                        // indicated element type

                        TypesElement typesElement = def.getTypesElement();

                        if (typesElement == null
                                || !typeValidator.validateType(typesElement,
                                        SchemaComponent.Type.ELEMENT_DECLARATION, messagePart.getElement())) {
                            validationMessage.append("Part '"
                                    + messagePart.getName()
                                    + "' defined in message '"
                                    + headerMessageName
                                    + "' mapped to soap:header in output for binding operation '"
                                    + bindingOperation.getName()
                                    + "' is invalid. No element declaration found for"
                                    + " element type "
                                    + messagePart.getElement()
                                    + ". A QName reference to a Schema component in a DESCRIPTION MUST"
                                    + " use the namespace defined in the targetNamespace attribute on"
                                    + " the xsd:schema element, or to a namespace defined in the"
                                    + " namespace attribute on an xsd:import element within the"
                                    + " xsd:schema element."
                                    + " (R2102 ws-i basic profile 1.1)");
                        }

                    }

                }
                log.trace("validateBindingInput end");
                return soapBodyType;
            }

        }

        private static class BindingOperationValidator {

            public QName validate(DefinitionsElement def, BindingElement binding, PortTypeElement portType, BindingOperationElement bindingOperation, ValidationInfo validationInfo) {

                log.trace("validate begin");

                ValidationMessage validationMessage = validationInfo.getValidationMessage();

                String bindingOperationName = bindingOperation.getName();

                SOAPOperationElement soapOperationElement = bindingOperation.getSOAPOperationElement();
                if (soapOperationElement == null) {
                    validationMessage.append("Binding operation "
                            + bindingOperationName
                            + " has no soap:operation element");
                } else {
                    soapOperationValidator.validate(bindingOperation, soapOperationElement, validationInfo);
                }

                // retrieve the operation to which this concrete operation is
                // bound, ie. the corresponding operation in portType whose
                // existence was verified by binding validator

                PortTypeOperationElement portTypeOperation = portType.getOperationElements().get(
                        bindingOperationName);

                QName soapBodyType = null;

                BindingOperationInputElement bindingOperationInput = bindingOperation.getInputElement();
                if (bindingOperationInput == null) {
                    validationMessage.append("Binding operation' "
                            + bindingOperationName
                            + "' has no wsdl:input element");
                } else {

                    soapBodyType = bindingInputValidator.validate(def, bindingOperation, portTypeOperation,
                            bindingOperationInput, validationInfo);

                }

                BindingOperationOutputElement bindingOperationOutput = bindingOperation.getOutputElement();
                if (bindingOperationOutput != null) {

                    bindingOutputValidator.validate(def, bindingOperation, portTypeOperation,
                            bindingOperationOutput, validationInfo);
                }

                for (BindingOperationFaultElement bindingFault : bindingOperation.getFaultElements().values()) {
                    bindingFaultValidator.validate(def, binding, bindingOperation, bindingFault, validationInfo);
                }

                log.trace("validate end");
                return soapBodyType;

            }

        }

        private static class BindingOutputValidator {

            public QName validate(DefinitionsElement def, BindingOperationElement bindingOperation, PortTypeOperationElement portTypeOperation, BindingOperationOutputElement bindingOutput, ValidationInfo validationInfo) {

                log.trace("validate begin");

                ValidationMessage validationMessage = validationInfo.getValidationMessage();

                SOAPBodyElement soapBody = bindingOutput.getSOAPBodyElement();
                Map<String, SOAPHeaderElement> soapHeaders = new HashMap<String, SOAPHeaderElement>();

                QName soapBodyType = null;

                if (soapBody == null) {
                    validationMessage.append("wsdl:output element in binding operation "
                            + bindingOperation.getName()
                            + " has no soap:body");
                    return null;
                } else {

                    if (portTypeOperation.getOutputElement() == null) {

                        validationMessage.append("No output defined for operation " + portTypeOperation.getName());
                        return null;

                    } else {

                        QName messageName = portTypeOperation.getOutputElement().getMessage();

                        // wsdl 1.1 (Section 3.7) in regards to headers, indicates "referenced
                        // message
                        // need not be the same as the message that defines the SOAP Body."
                        // this
                        // breaks
                        // bpel 1.1 semantics which associates a bpel variable with a single
                        // wsdl:message type. we will therefore dissallow this. a header part
                        // must
                        // ref
                        // same wsdl:message type used by soap body (which also allows us to
                        // index
                        // headers by part name)

                        for (SOAPHeaderElement soapHeader : bindingOutput.getSOAPHeaderElements()) {
                            if (soapHeader.getMessage() == null || !soapHeader.getMessage().equals(messageName)) {
                                validationMessage.append("soap:header in binding:output for binding operation '"
                                        + bindingOperation.getName()
                                        + "' references a message other than 'output' message type "
                                        + messageName
                                        + ". Configuration is not supported.");
                            } else {
                                if (soapHeaders.containsKey(soapHeader.getPart())) {
                                    validationMessage
                                            .append("soap:header in binding:output for binding operation '"
                                                    + bindingOperation.getName()
                                                    + "' binds abstract part '"
                                                    + soapHeader.getPart()
                                                    + "' which is already bound by a header.");
                                } else {
                                    soapHeaders.put(soapHeader.getPart(), soapHeader);
                                }
                            }
                        }

                    }

                    // now we validate that binding output is ws-i compliant. see below
                    // for comments

                    soapBodyType = validateBindingOutput(def, bindingOperation, portTypeOperation, soapBody,
                            soapHeaders, validationInfo);
                }

                log.trace("validate end");
                return soapBodyType;
            }

            private QName validateBindingOutput(DefinitionsElement def, BindingOperationElement bindingOperation, PortTypeOperationElement portTypeOperation, SOAPBodyElement soapBody, Map<String, SOAPHeaderElement> soapHeaders, ValidationInfo validationInfo) {

                log.trace("validateBindingOutput begin");

                ValidationMessage validationMessage = validationInfo.getValidationMessage();

                // validate the soapBody. for a doc-lit type binding if the part attribute on
                // soap:body is absent, then corresponding message may define zero or one part
                // each of which must use the 'element' attribute. if binding is an rpc-lit
                // binding, then zero or more parts may be specified, all of which must
                // reference
                // parts within message using the 'type' attribute. in either case, an empty
                // soap
                // body is valid, ie. no parts defined in either soap:body or wsdl:message.

                if (soapBody.getUse() == null || soapBody.getUse().equals(ENCODED_USE)) {
                    validationMessage.append("soap:body element in wsdl:output element in binding operation "
                            + bindingOperation.getName()
                            + " is invalid."
                            + " A wsdl:binding in a DESCRIPTION MUST use the value of"
                            + " 'literal' for the use attribute in all soapbind:body,"
                            + " soapbind:fault, soapbind:header and soapbind:headerfault"
                            + " elements."
                            + " (R2706 ws-i basic profile 1.1)");
                } else {
                    // spec allows a default value of 'literal' (see R2707)
                }

                // retrieve message element referenced by port type output element. we know
                // it exists because it was resolved by port type output validator

                MessageElement message = null;
                QName messageName = portTypeOperation.getOutputElement().getMessage();
                if (messageName.getNamespaceURI().equals(def.getTargetNamespace())) {
                    message = def.getMessageElements().get(messageName.getLocalPart());
                } else {
                    ImportElement importElement = def.getImportElements().get(messageName.getNamespaceURI());
                    if (importElement != null) {
                        DefinitionsElement temp = null;
                        try {
                            temp = importElement.getDocument().getDefinitionsElement();
                        } catch (WSDLException fatchance) {
                        }
                        message = temp.getMessageElements().get(messageName.getLocalPart());
                    }
                }

                Map<String, MessagePartElement> messagePartsMap = message.getPartElements();

                // validate concrete parts within doc-lit binding against the parts
                // within abstract message definition

                QName soapBodyType = null;

                if (validationInfo.getBindingStyle() == BindingStyle.DOCUMENT_STYLE) {

                    List<String> bodyParts = soapBody.getParts();

                    if (soapBody.getNamespace() != null) {
                        validationMessage.append("soap:body element in binding:operation "
                                + bindingOperation.getName()
                                + " is invalid."
                                + " A document-literal binding in a DESCRIPTION MUST NOT have"
                                + " the namespace attribute specified on contained soapbind:body,"
                                + " soapbind:header, soapbind:headerfault and soapbind:fault elements."
                                + " (R2716 ws-i basic profile 1.1)");
                    }

                    QName partType = null;
                    MessagePartElement messagePart = null;
                    if (bodyParts.size() == 0) {
                        if (messagePartsMap.size() > 1) {
                            validationMessage.append("soap:body element in binding:operation "
                                    + bindingOperation.getName()
                                    + " is invalid."
                                    + " If a document-literal binding in a DESCRIPTION does not"
                                    + " specify the parts attribute on a soapbind:body element, the"
                                    + " corresponding abstract wsdl:message MUST define zero or one"
                                    + " wsdl:parts."
                                    + " (R2210 ws-i basic profile 1.1)");
                        }
                        if (messagePartsMap.values().size() == 1) {
                            messagePart = messagePartsMap.values().iterator().next();
                            partType = messagePart.getElement();
                            if (partType == null) {
                                validationMessage.append("wsdl:part '"
                                        + messagePart.getName()
                                        + "' referenced by operation '"
                                        + bindingOperation.getName()
                                        + "' in doc-lit binding is invalid."
                                        + " A document-literal binding in a DESCRIPTION MUST refer, in"
                                        + " each of its soapbind:body element(s), only to wsdl:part"
                                        + " element(s) that have been defined using the element attribute."
                                        + " (R2204 ws-i basic profile 1.1)");
                            }
                        }
                    } else if (bodyParts.size() > 1) {
                        validationMessage.append("soap:body element in wsdl:output element in binding operation '"
                                + bindingOperation.getName()
                                + "' is invalid."
                                + " A document-literal binding in a DESCRIPTION MUST, in each of its"
                                + " soapbind:body element(s), have at most one part listed in the parts"
                                + " attribute, if the parts attribute is specified."
                                + " (R2201 ws-i basic profile 1.1)");

                    } else {
                        String partName = bodyParts.iterator().next();
                        messagePart = messagePartsMap.get(partName);
                        if (messagePart == null) {
                            validationMessage.append("wsdl:output message '"
                                    + messageName
                                    + "' contains no part with name matching name '"
                                    + partName
                                    + "' in corresponding soap:body");
                        } else {
                            partType = messagePart.getElement();
                            if (partType == null) {
                                validationMessage.append("wsdl:part '"
                                        + messagePart.getName()
                                        + "' within output message "
                                        + messageName
                                        + ", referenced by operation '"
                                        + bindingOperation.getName()
                                        + "' in doc-lit binding is invalid."
                                        + " A document-literal binding in a DESCRIPTION MUST refer, in"
                                        + " each of its soapbind:body element(s), only to wsdl:part"
                                        + " element(s) that have been defined using the element attribute."
                                        + " (R2204 ws-i basic profile 1.1)");
                            }
                        }
                    }

                    if (partType != null) {

                        // verify that an element decl exists within schema types for
                        // indicated element type

                        TypesElement typesElement = def.getTypesElement();

                        if (typesElement == null
                                || !typeValidator.validateType(typesElement,
                                        SchemaComponent.Type.ELEMENT_DECLARATION, partType)) {
                            validationMessage.append("wsdl:part '"
                                    + messagePart.getName()
                                    + "' within output message "
                                    + messageName
                                    + ", referenced by operation '"
                                    + bindingOperation.getName()
                                    + "' in doc-lit binding is invalid. No schema found for element "
                                    + partType
                                    + ". A QName reference to a Schema component in a DESCRIPTION MUST"
                                    + " use the namespace defined in the targetNamespace attribute on"
                                    + " the xsd:schema element, or a namespace defined on "
                                    + " namespace attribute on an xsd:import element within the"
                                    + " xsd:schema element."
                                    + " (R2102 ws-i basic profile 1.1)");
                        }

                        soapBodyType = partType;

                    }
                }

                // validate concrete parts within rpc-lit binding against the parts
                // within abstract message definition. the only thing we're looking for
                // here is that if parts are defined, they exist in message and are
                // typed with a 'type' attribute value, ie. not 'element'

                if (validationInfo.getBindingStyle() == BindingStyle.RPC_STYLE) {

                    List<String> bodyParts = soapBody.getParts();

                    if (soapBody.getNamespace() == null) {
                        validationMessage.append("soap:body element in binding:operation "
                                + bindingOperation.getName()
                                + " is invalid."
                                + " An rpc-literal binding in a DESCRIPTION MUST have the namespace"
                                + " attribute specified, the value of which MUST be an absolute URI,"
                                + " on contained soapbind:body elements."
                                + " (R2717 ws-i basic profile 1.1)");
                    }

                    if (bodyParts.size() > 0) {
                        if (messagePartsMap == null || messagePartsMap.size() == 0) {
                            validationMessage.append("wsdl:output message '"
                                    + messageName
                                    + "' contains no parts, but soap:body within output of"
                                    + " soap:operation "
                                    + bindingOperation.getName()
                                    + " specifies at least one part");
                        } else {
                            QName partType = null;
                            MessagePartElement messagePart = null;
                            for (String partName : bodyParts) {
                                messagePart = messagePartsMap.get(partName);
                                if (messagePart == null) {
                                    validationMessage.append("wsdl:output message '"
                                            + messageName
                                            + "' contains no part with name matching name '"
                                            + partName
                                            + "' in corresponding soap:body");
                                } else {
                                    partType = messagePart.getType();
                                    if (partType == null) {
                                        validationMessage.append("wsdl:part '"
                                                + messagePart.getName()
                                                + "' referenced by operation '"
                                                + bindingOperation.getName()
                                                + "' in rpc-lit binding is invalid."
                                                + " A rpc-literal binding in a DESCRIPTION MUST refer, in"
                                                + " each of its soapbind:body element(s), only to wsdl:part"
                                                + " element(s) that have been defined using the type attribute."
                                                + " (R2203 ws-i basic profile 1.1)");
                                    } else {

                                        // verify that an xml type decl exists within schema
                                        // types for indicated xml type, unless type is a
                                        // built-in
                                        // type in which case system is required to load schema
                                        // by
                                        // default

                                        if (!partType.getNamespaceURI().equals(XMLSchema.TYPE)) {

                                            TypesElement typesElement = def.getTypesElement();

                                            if (typesElement == null
                                                    || !typeValidator.validateType(typesElement,
                                                            SchemaComponent.Type.TYPE_DEFINITION, partType)) {
                                                validationMessage
                                                        .append("wsdl:part '"
                                                                + messagePart.getName()
                                                                + "' within output message "
                                                                + messageName
                                                                + ", referenced by operation '"
                                                                + bindingOperation.getName()
                                                                + "' in rpc-lit binding is invalid. No schema found for type "
                                                                + partType
                                                                + ". A QName reference to a Schema component in a DESCRIPTION MUST"
                                                                + " use the namespace defined in the targetNamespace attribute on"
                                                                + " the xsd:schema element, or to a namespace defined in the"
                                                                + " namespace attribute on an xsd:import element within the"
                                                                + " xsd:schema element."
                                                                + " (R2102 ws-i basic profile 1.1)");
                                            }

                                        }
                                    }

                                }
                            }
                        }

                    }
                }

                // validate the headers defined for binding output, specifically whether
                // they are mapped to a part in a message and that the part is typed
                // using the element attribute. this is required regardless of binding
                // type, ie rpc and document type

                for (SOAPHeaderElement soapHeader : soapHeaders.values()) {

                    if (soapHeader.getNamespace() != null
                            && validationInfo.getBindingStyle() == BindingStyle.RPC_STYLE) {
                        validationMessage.append("soap:header element in binding:operation "
                                + bindingOperation.getName()
                                + " is invalid."
                                + " An rpc-literal binding in a DESCRIPTION MUST NOT have the namespace"
                                + " attribute specified on contained soapbind:header, soapbind:headerfault"
                                + " and soapbind:fault elements."
                                + " (R2726 ws-i basic profile 1.1)");
                    } else if (soapHeader.getNamespace() != null) {
                        validationMessage.append("soap:header element in binding:operation "
                                + bindingOperation.getName()
                                + " is invalid."
                                + " A document-literal binding in a DESCRIPTION MUST NOT have the namespace"
                                + " attribute specified on contained soapbind:body, soapbind:header,"
                                + " soapbind:headerfault and soapbind:fault elements."
                                + " (R2716 ws-i basic profile 1.1)");
                    }

                    if (soapHeader.getUse() == null || soapHeader.getUse().equals(ENCODED_USE)) {
                        validationMessage.append("soap:header defined in output for binding operation '"
                                + bindingOperation.getName()
                                + "' is invalid."
                                + " A wsdl:binding in a DESCRIPTION MUST use the value of"
                                + " 'literal' for the use attribute in all soapbind:body,"
                                + " soapbind:fault, soapbind:header and soapbind:headerfault"
                                + " elements."
                                + " (R2706 ws-i basic profile 1.1)");
                    }

                    QName headerMessageName = soapHeader.getMessage();

                    MessagePartElement messagePart = messagePartsMap.get(soapHeader.getPart());

                    if (messagePart == null) {
                        validationMessage.append("No part found in message '"
                                + headerMessageName
                                + "' matching part name '"
                                + soapHeader.getPart()
                                + "' referenced by soap:header in output for binding operation '"
                                + bindingOperation.getName()
                                + "'.");
                    }
                    if (messagePart.getElement() == null) {
                        validationMessage.append("Part '"
                                + messagePart.getName()
                                + "' defined in message '"
                                + headerMessageName
                                + "' mapped to soap:header in output for binding operation '"
                                + bindingOperation.getName()
                                + "' is invalid."
                                + " A wsdl:binding in a DESCRIPTION MUST refer, in each of its"
                                + " soapbind:header, soapbind:headerfault and soapbind:fault"
                                + " elements, only to wsdl:part element(s) that have been defined"
                                + " using the element attribute."
                                + " (R2205 ws-i basic profile 1.1)");
                    } else {

                        // verify that an element decl exists within schema types for
                        // indicated element type

                        TypesElement typesElement = def.getTypesElement();

                        if (typesElement == null
                                || !typeValidator.validateType(typesElement,
                                        SchemaComponent.Type.ELEMENT_DECLARATION, messagePart.getElement())) {
                            validationMessage.append("Part '"
                                    + messagePart.getName()
                                    + "' defined in message '"
                                    + headerMessageName
                                    + "' mapped to soap:header in output for binding operation '"
                                    + bindingOperation.getName()
                                    + "' is invalid. No element declaration found for element "
                                    + messagePart.getElement()
                                    + ". A QName reference to a Schema component in a DESCRIPTION MUST"
                                    + " use the namespace defined in the targetNamespace attribute on"
                                    + " the xsd:schema element, or to a namespace defined in the"
                                    + " namespace attribute on an xsd:import element within the"
                                    + " xsd:schema element."
                                    + " (R2102 ws-i basic profile 1.1)");
                        }

                    }

                }
                log.trace("validateBindingOutput end");
                return soapBodyType;
            }

        }

        private static class BindingValidator {

            public void validate(DefinitionsElement def, BindingElement binding, ValidationInfo validationInfo) {

                log.trace("validate begin");

                // only soap/http bindings are indicated by ws-i basic profile. note that
                // 'callback' portTypes and bindings for asynchronous services, are not
                // validated as the callback mechanism is facilitated by ws-addressing

                ValidationMessage validationMessage = validationInfo.getValidationMessage();

                QName bindingName = new QName(def.getTargetNamespace(), binding.getName());

                SOAPBindingElement soapBindingElement = binding.getSOAPBindingElement();

                if (soapBindingElement == null) {
                    validationMessage.append("Binding "
                            + bindingName
                            + " is not a soap:binding. "
                            + " A wsdl:binding element in a DESCRIPTION MUST use WSDL SOAP"
                            + " Binding as defined in WSDL 1.1 Section 3."
                            + " (R2401 ws-i basic profile 1.1)");
                } else {
                    soapBindingValidator.validate(soapBindingElement, validationInfo);
                }

                QName portTypeName = binding.getType();
                PortTypeElement portType = null;

                if (portTypeName.getNamespaceURI().equals(def.getTargetNamespace())) {
                    portType = def.getPortTypeElements().get(portTypeName.getLocalPart());
                } else {
                    ImportElement importElement = def.getImportElements().get(portTypeName.getNamespaceURI());
                    if (importElement != null) {
                        DefinitionsElement temp = null;
                        try {
                            temp = importElement.getDocument().getDefinitionsElement();
                        } catch (WSDLException we) {
                            validationInfo.getValidationMessage().append(
                                    "Failed to resolve referenced portType "
                                            + portTypeName
                                            + " defined on binding "
                                            + bindingName
                                            + ". "
                                            + we);
                            return;
                        }
                        portType = temp.getPortTypeElements().get(portTypeName.getLocalPart());
                    }
                }

                if (portType == null) {
                    validationInfo.getValidationMessage().append(
                            "Failed to resolve referenced portType "
                                    + portTypeName
                                    + " defined on binding "
                                    + bindingName
                                    + ". ");
                    return;
                }

                portTypeValidator.validate(def, portType, validationInfo);

                // if port type is invalid, we cannot procede
                if (validationInfo.getValidationMessage().length() > 0) {
                    return;
                }

                Map<String, PortTypeOperationElement> portTypeOperations = portType.getOperationElements();
                Map<String, BindingOperationElement> bindingOperations = binding.getOperationElements();

                // verify that portType has same set of operations as binding
                if (portTypeOperations.size() != bindingOperations.size()) {
                    validationMessage.append("Binding "
                            + bindingName
                            + " binds to portType "
                            + portTypeName
                            + " incorrectly. A wsdl:binding in a DESCRIPTION MUST have the same"
                            + " set of wsdl:operations as the wsdl:portType to which it refers."
                            + " (R2718 ws-i basic profile 1.1)");
                } else {
                    Iterator<?> opitr = portTypeOperations.keySet().iterator();
                    while (opitr.hasNext()) {
                        String opname = (String)opitr.next();
                        if (bindingOperations.get(opname) == null) {
                            validationMessage.append("Operation '"
                                    + opname
                                    + "' not defined on binding "
                                    + bindingName
                                    + " bound to portType "
                                    + portTypeName
                                    + ". A wsdl:binding in a DESCRIPTION MUST have the same"
                                    + " set of wsdl:operations as the wsdl:portType to which it refers."
                                    + " (R2718 ws-i basic profile 1.1)");
                        }
                    }
                    opitr = bindingOperations.keySet().iterator();
                    while (opitr.hasNext()) {
                        String opname = (String)opitr.next();
                        if (portTypeOperations.get(opname) == null) {
                            validationMessage.append("Operation '"
                                    + opname
                                    + "' not defined on portType "
                                    + portTypeName
                                    + " referenced by binding "
                                    + bindingName
                                    + ". A wsdl:binding in a DESCRIPTION MUST have the same"
                                    + " set of wsdl:operations as the wsdl:portType to which it refers."
                                    + " (R2718 ws-i basic profile 1.1)");
                        }
                    }
                }

                // if concrete and abstract operations are out of sync, we cannot
                // proceed

                if (validationInfo.getValidationMessage().length() > 0) {
                    return;
                }

                // now we iterate over each operation and validate against profile.
                // note that although we may only invoke one operation, we are
                // validating that the service is compliant, which includes all
                // operations

                List<QName> soapBodyTypes = new ArrayList<QName>();
                for (BindingOperationElement bindingOperation : bindingOperations.values()) {
                    QName soapBodyType = null;
                    soapBodyType = bindingOperationValidator.validate(def, binding, portType, bindingOperation,
                            validationInfo);
                    if (soapBodyType != null) {
                        soapBodyTypes.add(soapBodyType);
                    } else {
                        // if other validation errors were found, or binding
                        // is rpc-lit, null is returned
                    }
                }

                // if binding is doc-lit, we verify that operation signatures are
                // unique. for rpc-lit bindings, this is not an issue, because op
                // name is used as a wrapper element and we know op name is unique

                List<QName> copy = new ArrayList<QName>(soapBodyTypes);
                if (validationInfo.getBindingStyle() == BindingStyle.DOCUMENT_STYLE) {
                    for (QName soapBodyType : soapBodyTypes) {
                        copy.remove(soapBodyType);
                        if (copy.contains(soapBodyType)) {
                            validationMessage.append("Operation signatures on binding "
                                    + bindingName
                                    + " are not unique. Element "
                                    + soapBodyType
                                    + " is payload type for more than one operation's input message."
                                    + " The operations in a wsdl:binding in a DESCRIPTION"
                                    + " MUST result in operation signatures that are different"
                                    + " from one another"
                                    + " (R2710 ws-i basic profile 1.1)");
                            break;
                        }
                    }

                }

                log.trace("validate end");

            }

        }

        private static class MessageValidator {

            public void validate(MessageElement message, ValidationInfo validationInfo) {

                log.trace("validate begin");

                ValidationMessage validationMessage = validationInfo.getValidationMessage();

                for (MessagePartElement partElement : message.getPartElements().values()) {
                    if (partElement.getElement() != null && partElement.getType() != null) {
                        validationMessage.append("wsdl:part '"
                                + partElement.getName()
                                + "' is invalid."
                                + " A wsdl:message in a DESCRIPTION MUST NOT specify both type and"
                                + " element attributes on the same wsdl:part."
                                + " (R2306 ws-i basic profile 1.1)");
                    }
                    if (partElement.getElement() != null) {
                        boolean isGED = true;
                        String namespaceURI = partElement.getElement().getNamespaceURI();
                        if (namespaceURI == null || namespaceURI.equals("")) {
                            isGED = false;
                        } else {
                            if (namespaceURI.equals(XMLSchema.TYPE)) {
                                isGED = false;
                            } else {
                                // we can only assume it's defined within a schema
                                // contained within types section
                            }
                        }
                        if (!isGED) {
                            validationMessage.append("wsdl:part '"
                                    + partElement.getName()
                                    + "' is invalid."
                                    + " A wsdl:message in a DESCRIPTION containing a wsdl:part that"
                                    + " uses the element attribute MUST refer, in that attribute, to"
                                    + " a global element declaration."
                                    + " (R2206 ws-i basic profile 1.1)");
                        }
                    }
                }

                log.trace("validate end");
            }

        }

        private static class PortTypeFaultValidator {

            public void validate(DefinitionsElement def, PortTypeOperationFaultElement faultElement, ValidationInfo validationInfo) {

                log.trace("validate begin");

                QName messageName = faultElement.getMessage();

                if (messageName == null) {
                    validationInfo.getValidationMessage().append("No message attribute defined on fault element.");
                    return;
                }

                MessageElement message = null;
                if (messageName.getNamespaceURI().equals(def.getTargetNamespace())) {
                    message = def.getMessageElements().get(messageName.getLocalPart());
                } else {
                    ImportElement importElement = def.getImportElements().get(messageName.getNamespaceURI());
                    if (importElement != null) {
                        DefinitionsElement temp = null;
                        try {
                            temp = importElement.getDocument().getDefinitionsElement();
                        } catch (WSDLException we) {
                            validationInfo.getValidationMessage().append(
                                    "Failed to resolve referenced message "
                                            + messageName
                                            + " defined on fault element. "
                                            + we);
                            return;
                        }
                        message = temp.getMessageElements().get(messageName.getLocalPart());
                    }
                }

                if (message == null) {
                    validationInfo.getValidationMessage().append(
                            "Failed to resolve referenced message " + messageName + " defined on fault element.");
                    return;
                }

                // validate abstract message
                messageValidator.validate(message, validationInfo);

                log.trace("validate end");
            }

        }

        private static class PortTypeInputValidator {

            public void validate(DefinitionsElement def, PortTypeOperationInputElement inputElement, ValidationInfo validationInfo) {

                log.trace("validate begin");

                QName messageName = inputElement.getMessage();

                if (messageName == null) {
                    validationInfo.getValidationMessage().append("No message attribute defined on input element.");
                    return;
                }

                MessageElement message = null;
                if (messageName.getNamespaceURI().equals(def.getTargetNamespace())) {
                    message = def.getMessageElements().get(messageName.getLocalPart());
                } else {
                    ImportElement importElement = def.getImportElements().get(messageName.getNamespaceURI());
                    if (importElement != null) {
                        DefinitionsElement temp = null;
                        try {
                            temp = importElement.getDocument().getDefinitionsElement();
                        } catch (WSDLException we) {
                            validationInfo.getValidationMessage().append(
                                    "Failed to resolve referenced message "
                                            + messageName
                                            + " defined on input element. "
                                            + we);
                            return;
                        }
                        message = temp.getMessageElements().get(messageName.getLocalPart());
                    }
                }

                if (message == null) {
                    validationInfo.getValidationMessage().append(
                            "Failed to resolve referenced message " + messageName + " defined on input element.");
                    return;
                }

                // validate abstract message
                messageValidator.validate(message, validationInfo);

                log.trace("validate end");
            }

        }

        private static class PortTypeOperationValidator {

            public void validate(DefinitionsElement def, PortTypeOperationElement operationElement, ValidationInfo validationInfo) {

                log.trace("validate begin");

                // the optional parameterOrder attribute is not relevant as it is used
                // by code generators, ie. axis wsdl2java tool, to build java objects
                // from an rpc binding and can be safely ignored

                ValidationMessage validationMessage = validationInfo.getValidationMessage();

                // 'output/input' MEP's are not supported, ie. a portType that has
                // an operation whose first element is output is disallowed. the
                // notification (output) pattern should instead be modeled as a
                // separate 'callback' portType with a single input operation.
                // output/input pattern is flipped and becomes request/response
                // (input/output) despite fact that server is initiating the
                // message exchange

                if (operationElement.getStyle() == Style.NOTIFICATION
                        || operationElement.getStyle() == Style.SOLICIT_RESPONSE) {
                    validationMessage.append("Invalid operation "
                            + operationElement.getName()
                            + ". A DESCRIPTION MUST NOT use Solicit-Response and"
                            + " Notification type operations in a wsdl:portType definition."
                            + " (R2303 ws-i basic profile 1.1)");
                }

                // validate input
                portTypeInputValidator.validate(def, operationElement.getInputElement(), validationInfo);

                // validate output
                if (operationElement.getOutputElement() != null) {
                    portTypeOutputValidator.validate(def, operationElement.getOutputElement(), validationInfo);

                }

                // validate faults
                for (PortTypeOperationFaultElement faultElement : operationElement.getFaultElements().values()) {
                    portTypeFaultValidator.validate(def, faultElement, validationInfo);

                }

                log.trace("validate end");
            }

        }

        private static class PortTypeOutputValidator {

            public void validate(DefinitionsElement def, PortTypeOperationOutputElement outputElement, ValidationInfo validationInfo) {

                log.trace("validate begin");

                QName messageName = outputElement.getMessage();

                if (messageName == null) {
                    validationInfo.getValidationMessage()
                            .append("No message attribute defined on output element.");
                    return;
                }

                MessageElement message = null;
                if (messageName.getNamespaceURI().equals(def.getTargetNamespace())) {
                    message = def.getMessageElements().get(messageName.getLocalPart());
                } else {
                    ImportElement importElement = def.getImportElements().get(messageName.getNamespaceURI());
                    if (importElement != null) {
                        DefinitionsElement temp = null;
                        try {
                            temp = importElement.getDocument().getDefinitionsElement();
                        } catch (WSDLException we) {
                            validationInfo.getValidationMessage().append(
                                    "Failed to resolve referenced message "
                                            + messageName
                                            + " defined on output element. "
                                            + we);
                            return;
                        }
                        message = temp.getMessageElements().get(messageName.getLocalPart());
                    }
                }

                if (message == null) {
                    validationInfo.getValidationMessage().append(
                            "Failed to resolve referenced message " + messageName + " defined on output element.");
                    return;
                }

                // validate abstract message
                messageValidator.validate(message, validationInfo);

                log.trace("validate end");
            }

        }

        private static class PortTypeValidator {

            public void validate(DefinitionsElement def, PortTypeElement portType, ValidationInfo validationInfo) {

                log.trace("validate begin");

                // operation names must be unique within a portType. no overloading
                // allowed by profile. this was validated within parser due to the
                // way parent element indexes ops, i.e. by name

                for (PortTypeOperationElement operationElement : portType.getOperationElements().values()) {
                    portTypeOperationValidator.validate(def, operationElement, validationInfo);
                }

                log.trace("validate end");

            }

        }

        private static class ServiceValidator {

            public void validate(DefinitionsElement def, ServiceElement service, ValidationInfo validationInfo) {
                log.trace("validate begin");
                for (ServicePortElement port : service.getPortElements().values()) {
                    validatePort(def, port, validationInfo);
                }
                log.trace("validate end");
            }

            private void validatePort(DefinitionsElement def, ServicePortElement port, ValidationInfo validationInfo) {

                ValidationMessage validationMessage = validationInfo.getValidationMessage();

                SOAPAddressElement soapAddress = port.getSOAPAddressElement();

                if (soapAddress == null) {
                    validationMessage.append("No soap:address found in port " + port.getName());
                } else {
                    String locationURI = soapAddress.getLocation();
                    if (locationURI == null) {
                        validationMessage.append("invalid location attribute in " + "soap:address");
                    } else {
                        // uri may contain variables which will be bound by transport,
                        // i.e. we can't test to see if it's a valid URL
                    }
                }

                QName bindingName = port.getBinding();
                BindingElement binding = null;

                if (bindingName.getNamespaceURI().equals(def.getTargetNamespace())) {
                    binding = def.getBindingElements().get(bindingName.getLocalPart());
                } else {
                    ImportElement importElement = def.getImportElements().get(bindingName.getNamespaceURI());
                    if (importElement != null) {
                        DefinitionsElement temp = null;
                        try {
                            temp = importElement.getDocument().getDefinitionsElement();
                        } catch (WSDLException we) {
                            validationInfo.getValidationMessage().append(
                                    "Failed to resolve reference to binding '"
                                            + bindingName
                                            + " defined on port '"
                                            + port.getName()
                                            + "'. "
                                            + we);
                            return;
                        }
                        binding = temp.getBindingElements().get(bindingName.getLocalPart());
                    }
                }

                if (binding == null) {
                    validationMessage.append("Failed to resolve reference to binding '"
                            + bindingName
                            + " defined on port '"
                            + port.getName()
                            + "'.");
                } else {
                    bindingValidator.validate(def, binding, validationInfo);
                }

            }

        }

        private static class SOAPBindingValidator {

            public void validate(SOAPBindingElement soapBindingElement, ValidationInfo validationInfo) {

                log.trace("validate begin");

                ValidationMessage validationMessage = validationInfo.getValidationMessage();

                String transport = soapBindingElement.getTransport();
                if (transport == null || !(transport.equals(SOAP_HTTP_TRANSPORT))) {
                    validationMessage
                            .append("A wsdl:binding element in a DESCRIPTION MUST specify the HTTP transport"
                                    + " protocol with SOAP binding. Specifically, the transport attribute"
                                    + " of its soapbind:binding child MUST have the value "
                                    + " \"http://schemas.xmlsoap.org/soap/http\"."
                                    + " (R2702 ws-i basic profile 1.1)");
                }

                // profile indicates that if style is not specified then we assume 'document'
                // otherwise it is as reads, ie. 'document' or 'rpc'. note that we assume the
                // binding is either a rpc_lit or a doc_lit binding at this point. other steps
                // will be taken to further validate this in accordance with profile
                // definition.
                // if the definition fails validation for any reason, then style is invaled
                // and should be ignored

                String style = soapBindingElement.getStyle();
                if (style != null && style.equals(RPC_BINDING_STYLE)) {
                    validationInfo.setBindingStyle(BindingStyle.RPC_STYLE);
                } else {
                    validationInfo.setBindingStyle(BindingStyle.DOCUMENT_STYLE);
                }

                log.trace("validate end");
            }

        }

        private static class SOAPOperationValidator {

            public void validate(BindingOperationElement bindingOperation, SOAPOperationElement soapOperationElement, ValidationInfo validationInfo) {

                log.trace("validate begin");

                ValidationMessage validationMessage = validationInfo.getValidationMessage();

                String operationStyle = soapOperationElement.getStyle();
                if (operationStyle != null) {
                    if (operationStyle.equals(RPC_BINDING_STYLE)) {
                        if (validationInfo.getBindingStyle() == BindingStyle.DOCUMENT_STYLE) {
                            validationMessage.append("soap:operation style attribute in binding operation '"
                                    + bindingOperation.getName()
                                    + "' is inconsistent with that of parent binding."
                                    + " A wsdl:binding in a DESCRIPTION MUST either be an rpc-literal"
                                    + " binding or a document-literal binding."
                                    + " (R2705 ws-i basic profile 1.1)");
                        }
                    } else {
                        if (validationInfo.getBindingStyle() == BindingStyle.RPC_STYLE) {
                            validationMessage.append("soap:operation style attribute in binding operation '"
                                    + bindingOperation.getName()
                                    + "' is inconsistent with that of parent binding."
                                    + " A wsdl:binding in a DESCRIPTION MUST either be an rpc-literal"
                                    + " binding or a document-literal binding."
                                    + " (R2705 ws-i basic profile 1.1)");
                        }
                    }
                }

                log.trace("validate end");
            }

        }

        private static class TypesValidator {

            public void validate(TypesElement typesElement, ValidationInfo validationInfo) {

                log.trace("validate begin");

                ValidationMessage validationMessage = validationInfo.getValidationMessage();

                List<Element> schemaElements = DOMUtil.getChildElements(typesElement, new QName(XMLSchema.TYPE,
                        "schema"));

                for (Element schemaElement : schemaElements) {

                    if (schemaElement.getAttribute("targetNamespace").equals("")) {

                        List<Element> childElements = DOMUtil.getChildElements(schemaElement, new QName("*", "*"));

                        for (Element childElement : childElements) {
                            if (childElement.getLocalName().equals("import")) {
                                continue;
                            }
                            if (childElement.getLocalName().equals("annotation")) {
                                continue;
                            }
                            validationMessage.append("All xsd:schema elements contained in a wsdl:types"
                                    + " element of a DESCRIPTION MUST have a targetNamespace attribute"
                                    + " with a valid and non-null value, UNLESS the xsd:schema element"
                                    + " has xsd:import and/or xsd:annotation as its only child element(s)."
                                    + " (R2105 ws-i basic profile 1.1)");
                        }

                    }
                }

                log.trace("validate end");
            }

        }

        private static class TypeValidator {

            /**
             * Validates that a schema with tns matching type is 'visible'. WSDL 1.1 visibility
             * rules allow a type to be referenced if defined within an inlined schema or if
             * imported by an inlined schema via an xs:import. Nested imports, i.e. types
             * imported by an imported schema , are NOT visible. Note that the actual reference
             * to referenced type is resolved at runtime, which allows for the fact that schema
             * may be defined within a separate deployment.
             * <p>
             * Note that basic profile is not clear regarding whether or not schemas within
             * types section of imported wsdl documents are visible to containing definition.
             * wsdl 20 is clear, however, on this point, i.e. they are NOT visible. We'll
             * adhere to this rule.
             * 
             * @param types
             * @param type
             * @param typeName
             * @return boolean true if schema found, false if not.
             * @throws WSDLException
             */
            public boolean validateType(TypesElement typesElement, SchemaComponent.Type type, QName typeName) {
                log.trace("validateType begin");

                List<Element> schemaElements = DOMUtil.getChildElements(typesElement, new QName(XMLSchema.TYPE,
                        "schema"));

                for (Element schemaElement : schemaElements) {

                    String targetNamespace = schemaElement.getAttribute("targetNamespace");
                    if (targetNamespace != null && targetNamespace.equals(typeName.getNamespaceURI())) {
                        return true;
                    } else {
                        List<Element> importElements = DOMUtil.getChildElements(schemaElement, new QName(
                                XMLSchema.TYPE, "import"));
                        for (Element importElement : importElements) {
                            String namespace = importElement.getAttribute("namespace");
                            if (namespace != null && namespace.equals(typeName.getNamespaceURI())) {
                                return true;
                            }
                        }
                    }

                }

                log.trace("validateType end");
                return false;

            }

        }

    }

    private static class DOMUtil {

        public static List<Element> getChildElements(Element parent, QName childName) {
            List<Element> answer = new ArrayList<Element>();
            NodeList nodeList = parent.getChildNodes();
            for (int i = 0; i < nodeList.getLength(); i++) {
                if (nodeList.item(i).getNodeType() != Node.ELEMENT_NODE) {
                    continue;
                }
                Element element = (Element)nodeList.item(i);
                if (childName.getNamespaceURI().equals(XMLConstants.NULL_NS_URI)) {
                    if (element.getNamespaceURI() != null) {
                        continue;
                    }
                } else {
                    if (element.getNamespaceURI() == null) {
                        continue;
                    }
                    if (!childName.getNamespaceURI().equals("*")
                            && !element.getNamespaceURI().equals(childName.getNamespaceURI())) {
                        continue;
                    }
                }
                if (!childName.getLocalPart().equals("*")
                        && !element.getLocalName().equals(childName.getLocalPart())) {
                    continue;
                }
                answer.add(element);
            }
            return answer;
        }

        public static Element getChildElement(Element parent, QName childName) {
            NodeList nodeList = parent.getChildNodes();
            for (int i = 0; i < nodeList.getLength(); i++) {
                if (nodeList.item(i).getNodeType() != Node.ELEMENT_NODE) {
                    continue;
                }
                Element element = (Element)nodeList.item(i);
                if (childName.getNamespaceURI().equals(XMLConstants.NULL_NS_URI)) {
                    if (element.getNamespaceURI() != null) {
                        continue;
                    }
                } else {
                    if (element.getNamespaceURI() == null) {
                        continue;
                    }
                    if (!childName.getNamespaceURI().equals("*")
                            && !element.getNamespaceURI().equals(childName.getNamespaceURI())) {
                        continue;
                    }
                }
                if (!childName.getLocalPart().equals("*")
                        && !element.getLocalName().equals(childName.getLocalPart())) {
                    continue;
                }
                return element;
            }
            return null;
        }

    }

    private static class PolicyUtil {

        public static ElementPolicy getAttachedPolicy(ComponentContext componentContext, DOMSource source) throws DeploymentException {
            ElementPolicy elementPolicy = null;
            for (PolicyFactory factory : SystemContext.getContext().getSystem().getPolicyFactories()) {
                ElementPolicy temp = factory.readElementPolicy(componentContext, source);
                if (elementPolicy == null) {
                    elementPolicy = temp;
                } else {
                    throw new DeploymentException("Attached policy expressions MUST be identically typed.");
                }
            }
            return elementPolicy;
        }

        public static void writePolicy(PolicyExpression policyExpression, Element parent) {
            parent.appendChild(parent.getOwnerDocument().importNode(policyExpression, true));
        }

    }

    private static class ValidationInfo {

        private ValidationMessage validationMessage = null;
        private BindingStyle bindingStyle;

        // if a validationMessage exists, then bindingFactory is invalid. the type
        // is assumed early in validation chain, ie. within SOAPBindingValidator,
        // but is dependent upon validation by other downstream validators as
        // well, ie. meaning that not all validators may have completed

        public BindingStyle getBindingStyle() {
            return bindingStyle;
        }

        public ValidationMessage getValidationMessage() {
            return validationMessage;
        }

        public void setBindingStyle(BindingStyle bindingStyle) {
            this.bindingStyle = bindingStyle;
        }

        public void setValidationMessage(ValidationMessage validationMessage) {
            this.validationMessage = validationMessage;
        }

    }

    private static class ValidationMessage {

        private StringBuffer sb = null;
        private int errorCount = 1;

        public void append(String message) {
            if (sb == null) {
                sb = new StringBuffer();
                sb.append((errorCount++) + ". " + wrapText(message));
            } else {
                sb.append(LINE_BREAK);
                sb.append((errorCount++) + ". " + wrapText(message));
            }
        }

        public int length() {
            if (sb != null) {
                return sb.length();
            } else {
                return 0;
            }
        }

        public String toString() {
            if (sb != null) {
                return sb.toString();
            } else {
                return null;
            }
        }

        private String wrapText(String unwrappedText) {
            StringBuffer wrappedText = new StringBuffer();
            int lineCharCount = 0;
            for (int i = 0; i < unwrappedText.length(); i++) {
                char currentChar = unwrappedText.charAt(i);
                if (lineCharCount > 80) {
                    if (Character.isWhitespace(currentChar)) {
                        wrappedText.append(LINE_BREAK);
                        wrappedText.append("   ");
                        lineCharCount = 0;
                    } else {
                        wrappedText.append(currentChar);
                    }
                } else {
                    lineCharCount++;
                    wrappedText.append(currentChar);
                }
            }
            return wrappedText.toString();
        }

    }

    /** ******************************* private methods ********************************** * */

    private void readTopLevelComponents(ComponentContext componentContext, WSDL11Document doc) throws DeploymentException {

        DefinitionsElement def = doc.getDefinitionsElement();

        typesDefinitionReader.readTypes(def, componentContext);

        policyDefinitionReader.readPolicy(def, componentContext);

        for (EngineElement engineElement : def.getEngineElements().values()) {
            String localPart = engineElement.getAttribute("name");
            if (localPart.contains(":")) {
                throw new DeploymentException("Top level name '" + localPart + "' must not be prefixed.");
            }
            ComponentName name = new ComponentName(def.getTargetNamespace(), localPart);
            engineDefinitionReader.readDefinition(def, componentContext, name);
        }

        for (ApplicationElement applicationElement : def.getApplicationElements().values()) {
            String localPart = applicationElement.getAttribute("name");
            if (localPart.contains(":")) {
                throw new DeploymentException("Top level name '" + localPart + "' must not be prefixed.");
            }
            ComponentName name = new ComponentName(def.getTargetNamespace(), localPart);
            applicationDefinitionReader.readDefinition(def, componentContext, name);
        }

        for (ServiceElement serviceElement : def.getServiceElements().values()) {
            String localPart = serviceElement.getAttribute("name");
            if (localPart.contains(":")) {
                throw new DeploymentException("Top level name '" + localPart + "' must not be prefixed.");
            }
            ComponentName name = new ComponentName(def.getTargetNamespace(), localPart);
            serviceDefinitionReader.readDefinition(def, componentContext, name);
        }

        for (BindingElement bindingElement : def.getBindingElements().values()) {
            String localPart = bindingElement.getAttribute("name");
            if (localPart.contains(":")) {
                throw new DeploymentException("Top level name '" + localPart + "' must not be prefixed.");
            }
            ComponentName name = new ComponentName(def.getTargetNamespace(), localPart);
            bindingDefinitionReader.readDefinition(def, componentContext, name);
        }

        for (PortTypeElement portTypeElement : def.getPortTypeElements().values()) {
            String localPart = portTypeElement.getAttribute("name");
            if (localPart.contains(":")) {
                throw new DeploymentException("Top level name '" + localPart + "' must not be prefixed.");
            }
            ComponentName name = new ComponentName(def.getTargetNamespace(), localPart);
            interfaceDefinitionReader.readDefinition(def, componentContext, name);
        }

        for (MessageElement messageElement : def.getMessageElements().values()) {
            String localPart = messageElement.getAttribute("name");
            if (localPart.contains(":")) {
                throw new DeploymentException("Top level name '" + localPart + "' must not be prefixed.");
            }
            ComponentName name = new ComponentName(def.getTargetNamespace(), localPart);
            messageDefinitionReader.readDefinition(def, componentContext, name);
        }

    }

}
