/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.service;

import org.bluestemsoftware.specification.eoa.component.policy.rt.ElementPolicy;
import org.bluestemsoftware.specification.eoa.component.service.BindingReference;
import org.bluestemsoftware.specification.eoa.component.service.Endpoint;

/**
 * WSDL 2.0 adjunct spec decorates an endpoint with type specific transport attributes, which
 * IMO should be configured via ws-policy instead, ...
 */
public class HTTPEndpoint extends Endpoint {

    private String authenticationScheme;
    private String authenticationRealm;

    public HTTPEndpoint(String name, BindingReference bindingReference, String address, ElementPolicy elementPolicy) {
        super(name, bindingReference, address, elementPolicy);
    }

    public HTTPEndpoint(String name, BindingReference bindingReference, String address, String authenticationScheme,
            String authenticationRealm, ElementPolicy elementPolicy) {
        super(name, bindingReference, address, elementPolicy);
        this.authenticationScheme = authenticationScheme;
        this.authenticationRealm = authenticationRealm;
    }

    public String getAuthenticationScheme() {
        return authenticationScheme;
    }

    public String getAuthenticationRealm() {
        return authenticationRealm;
    }

}
